"""Python API for the lab streaming layer.

The lab streaming layer provides a set of functions to make instrument data
accessible in real time within a lab network. From there, streams can be
picked up by recording programs, viewing programs or custom experiment
applications that access data streams in real time.

The API covers two areas:
- The "push API" allows to create stream outlets and to push data (regular
  or irregular measurement time series, event data, coded audio/video frames,
  etc.) into them.
- The "pull API" allows to create stream inlets and read time-synched
  experiment data from them (for recording, viewing or experiment control).

pylsl has been tested with Python 2.7 and 3.4.

"""

import os
import platform
import struct
from ctypes import CDLL, util, byref, c_char_p, c_void_p, c_double, c_int, \
    c_long, c_float, c_short, c_byte, c_longlong, cast, POINTER

__all__ = ['IRREGULAR_RATE', 'DEDUCED_TIMESTAMP', 'FOREVER', 'cf_float32',
           'cf_double64', 'cf_string', 'cf_int32', 'cf_int16', 'cf_int8',
           'cf_int64', 'cf_undefined', 'protocol_version', 'library_version',
           'library_info',
           'local_clock', 'StreamInfo', 'StreamOutlet', 'resolve_streams',
           'resolve_byprop', 'resolve_bypred', 'StreamInlet', 'XMLElement',
           'ContinuousResolver', 'TimeoutError', 'LostError',
           'InvalidArgumentError', 'InternalError', 'stream_info',
           'stream_outlet', 'stream_inlet', 'xml_element', 'timeout_error',
           'lost_error', 'vectorf', 'vectord', 'vectorl', 'vectori',
           'vectors', 'vectorc', 'vectorstr', 'resolve_stream']

# =================
# === Constants ===
# =================

# Constant to indicate that a stream has variable sampling rate.
IRREGULAR_RATE = 0.0

# Constant to indicate that a sample has the next successive time stamp
# according to the stream's defined sampling rate. Optional optimization to
# transmit less data per sample.
DEDUCED_TIMESTAMP = -1.0

# A very large time value (ca. 1 year); can be used in timeouts.
FOREVER = 32000000.0

# Value formats supported by LSL. LSL data streams are sequences of samples,
# each of which is a same-size vector of values with one of the below types.

# For up to 24-bit precision measurements in the appropriate physical unit (
# e.g., microvolts). Integers from -16777216 to 16777216 are represented
# accurately.
cf_float32 = 1
# For universal numeric data as long as permitted by network and disk budget.
#  The largest representable integer is 53-bit.
cf_double64 = 2
# For variable-length ASCII strings or data blobs, such as video frames,
# complex event descriptions, etc.
cf_string = 3
# For high-rate digitized formats that require 32-bit precision. Depends
# critically on meta-data to represent meaningful units. Useful for
# application event codes or other coded data.
cf_int32 = 4
# For very high bandwidth signals or CD quality audio (for professional audio
#  float is recommended).
cf_int16 = 5
# For binary signals or other coded data.
cf_int8 = 6
# For now only for future compatibility. Support for this type is not
# available on all languages and platforms.
cf_int64 = 7
# Can not be transmitted.
cf_undefined = 0

# Post processing flags
proc_none = 0  # No automatic post-processing; return the ground-truth time stamps for manual post-processing.
proc_clocksync = 1  # Perform automatic clock synchronization; equivalent to manually adding the time_correction().
proc_dejitter = 2  # Remove jitter from time stamps using a smoothing algorithm to the received time stamps.
proc_monotonize = 4  # Force the time-stamps to be monotonically ascending. Only makes sense if timestamps are dejittered.
proc_threadsafe = 8  # Post-processing is thread-safe (same inlet can be read from by multiple threads).
proc_ALL = proc_none | proc_clocksync | proc_dejitter | proc_monotonize | proc_threadsafe


# ==========================================================
# === Free Functions provided by the lab streaming layer ===
# ==========================================================

def protocol_version():
    """Protocol version.

    The major version is protocol_version() / 100;
    The minor version is protocol_version() % 100;

    Clients with different minor versions are protocol-compatible with each
    other while clients with different major versions will refuse to work
    together.

    """
    return lib.lsl_protocol_version()


def library_version():
    """Version of the underlying liblsl library.

    The major version is library_version() / 100;
    The minor version is library_version() % 100;

    """
    return lib.lsl_library_version()


def library_info():
    """Get a string containing library information. The format of the string shouldn't be used
    for anything important except giving a a debugging person a good idea which exact library
    version is used."""
    return lib.lsl_library_info().decode('utf-8')


def local_clock():
    """Obtain a local system time stamp in seconds.

    The resolution is better than a milisecond. This reading can be used to
    assign time stamps to samples as they are being acquired.

    If the "age" of a sample is known at a particular time (e.g., from USB
    transmission delays), it can be used as an offset to lsl_local_clock() to
    obtain a better estimate of when a sample was actually captured. See
    StreamOutlet.push_sample() for a use case.

    """
    return lib.lsl_local_clock()


# ==========================
# === Stream Declaration ===
# ==========================

class StreamInfo:
    """The StreamInfo object stores the declaration of a data stream.

    Represents the following information:
     a) stream data format (#channels, channel format)
     b) core information (stream name, content type, sampling rate)
     c) optional meta-data about the stream content (channel labels,
        measurement units, etc.)

    Whenever a program wants to provide a new stream on the lab network it will
    typically first create a StreamInfo to describe its properties and then
    construct a StreamOutlet with it to create the stream on the network.
    Recipients who discover the outlet can query the StreamInfo; it is also
    written to disk when recording the stream (playing a similar role as a file
    header).

    """

    def __init__(self, name='untitled', type='', channel_count=1,
                 nominal_srate=IRREGULAR_RATE, channel_format=cf_float32,
                 source_id='', handle=None):
        """Construct a new StreamInfo object.

        Core stream information is specified here. Any remaining meta-data can
        be added later.

        Keyword arguments:
        name -- Name of the stream. Describes the device (or product series)
                that this stream makes available (for use by programs,
                experimenters or data analysts). Cannot be empty.
        type -- Content type of the stream. By convention LSL uses the content
                types defined in the XDF file format specification where
                applicable (https://github.com/sccn/xdf). The content type is the
                preferred way to find streams (as opposed to searching by name).
        channel_count -- Number of channels per sample. This stays constant for
                         the lifetime of the stream. (default 1)
        nominal_srate -- The sampling rate (in Hz) as advertised by the data
                         source, regular (otherwise set to IRREGULAR_RATE).
                         (default IRREGULAR_RATE)
        channel_format -- Format/type of each channel. If your channels have
                          different formats, consider supplying multiple
                          streams or use the largest type that can hold
                          them all (such as cf_double64). It is also allowed
                          to pass this as a string, without the cf_ prefix,
                          e.g., 'float32' (default cf_float32)
        source_id -- Unique identifier of the device or source of the data, if
                     available (such as the serial number). This is critical
                     for system robustness since it allows recipients to
                     recover from failure even after the serving app, device or
                     computer crashes (just by finding a stream with the same
                     source id on the network again). Therefore, it is highly
                     recommended to always try to provide whatever information
                     can uniquely identify the data source itself.
                     (default '')

        """
        if handle is not None:
            self.obj = c_void_p(handle)
        else:
            if isinstance(channel_format, str):
                channel_format = string2fmt[channel_format]
            self.obj = lib.lsl_create_streaminfo(c_char_p(str.encode(name)),
                                                 c_char_p(str.encode(type)),
                                                 channel_count,
                                                 c_double(nominal_srate),
                                                 channel_format,
                                                 c_char_p(str.encode(source_id)))
            self.obj = c_void_p(self.obj)
            if not self.obj:
                raise RuntimeError("could not create stream description "
                                   "object.")

    def __del__(self):
        """ Destroy a previously created StreamInfo object. """
        # noinspection PyBroadException
        try:
            lib.lsl_destroy_streaminfo(self.obj)
        except:
            pass

    # === Core Information (assigned at construction) ===

    def name(self):
        """Name of the stream.

        This is a human-readable name. For streams offered by device modules,
        it refers to the type of device or product series that is generating
        the data of the stream. If the source is an application, the name may
        be a more generic or specific identifier. Multiple streams with the
        same name can coexist, though potentially at the cost of ambiguity (for
        the recording app or experimenter).

        """
        return lib.lsl_get_name(self.obj).decode('utf-8')

    def type(self):
        """Content type of the stream.

        The content type is a short string such as "EEG", "Gaze" which
        describes the content carried by the channel (if known). If a stream
        contains mixed content this value need not be assigned but may instead
        be stored in the description of channel types. To be useful to
        applications and automated processing systems using the recommended
        content types is preferred.

        """
        return lib.lsl_get_type(self.obj).decode('utf-8')

    def channel_count(self):
        """Number of channels of the stream.

        A stream has at least one channel; the channel count stays constant for
        all samples.

        """
        return lib.lsl_get_channel_count(self.obj)

    def nominal_srate(self):
        """Sampling rate of the stream, according to the source (in Hz).

        If a stream is irregularly sampled, this should be set to
        IRREGULAR_RATE.

        Note that no data will be lost even if this sampling rate is incorrect
        or if a device has temporary hiccups, since all samples will be
        transmitted anyway (except for those dropped by the device itself).
        However, when the recording is imported into an application, a good
        data importer may correct such errors more accurately if the advertised
        sampling rate was close to the specs of the device.

        """
        return lib.lsl_get_nominal_srate(self.obj)

    def channel_format(self):
        """Channel format of the stream.

        All channels in a stream have the same format. However, a device might
        offer multiple time-synched streams each with its own format.

        """
        return lib.lsl_get_channel_format(self.obj)

    def source_id(self):
        """Unique identifier of the stream's source, if available.

        The unique source (or device) identifier is an optional piece of
        information that, if available, allows that endpoints (such as the
        recording program) can re-acquire a stream automatically once it is
        back online.

        """
        return lib.lsl_get_source_id(self.obj).decode('utf-8')

    # === Hosting Information (assigned when bound to an outlet/inlet) ===

    def version(self):
        """Protocol version used to deliver the stream."""
        return lib.lsl_get_version(self.obj)

    def created_at(self):
        """Creation time stamp of the stream.

        This is the time stamp when the stream was first created
        (as determined via local_clock() on the providing machine).

        """
        return lib.lsl_get_created_at(self.obj)

    def uid(self):
        """Unique ID of the stream outlet instance (once assigned).

        This is a unique identifier of the stream outlet, and is guaranteed to
        be different across multiple instantiations of the same outlet (e.g.,
        after a re-start).

        """
        return lib.lsl_get_uid(self.obj).decode('utf-8')

    def session_id(self):
        """Session ID for the given stream.

        The session id is an optional human-assigned identifier of the
        recording session. While it is rarely used, it can be used to prevent
        concurrent recording activitites on the same sub-network (e.g., in
        multiple experiment areas) from seeing each other's streams
        (can be assigned in a configuration file read by liblsl, see also
        Network Connectivity in the LSL wiki).

        """
        return lib.lsl_get_session_id(self.obj).decode('utf-8')

    def hostname(self):
        """Hostname of the providing machine."""
        return lib.lsl_get_hostname(self.obj).decode('utf-8')

    # === Data Description (can be modified) ===

    def desc(self):
        """Extended description of the stream.

        It is highly recommended that at least the channel labels are described
        here. See code examples on the LSL wiki. Other information, such
        as amplifier settings, measurement units if deviating from defaults,
        setup information, subject information, etc., can be specified here, as
        well. Meta-data recommendations follow the XDF file format project
        (github.com/sccn/xdf/wiki/Meta-Data or web search for: XDF meta-data).

        Important: if you use a stream content type for which meta-data
        recommendations exist, please try to lay out your meta-data in
        agreement with these recommendations for compatibility with other
        applications.

        """
        return XMLElement(lib.lsl_get_desc(self.obj))

    def as_xml(self):
        """Retrieve the entire stream_info in XML format.

        This yields an XML document (in string form) whose top-level element is
        <description>. The description element contains one element for each
        field of the stream_info class, including:
        a) the core elements <name>, <type>, <channel_count>, <nominal_srate>,
           <channel_format>, <source_id>
        b) the misc elements <version>, <created_at>, <uid>, <session_id>,
           <v4address>, <v4data_port>, <v4service_port>, <v6address>,
           <v6data_port>, <v6service_port>
        c) the extended description element <desc> with user-defined
           sub-elements.

        """
        return lib.lsl_get_xml(self.obj).decode('utf-8')


# =====================
# === Stream Outlet ===
# =====================

class StreamOutlet:
    """A stream outlet.

    Outlets are used to make streaming data (and the meta-data) available on
    the lab network.

    """

    def __init__(self, info, chunk_size=0, max_buffered=360):
        """Establish a new stream outlet. This makes the stream discoverable.

        Keyword arguments:
        description -- The StreamInfo object to describe this stream. Stays
                constant over the lifetime of the outlet.
        chunk_size --- Optionally the desired chunk granularity (in samples)
                       for transmission. If unspecified, each push operation
                       yields one chunk. Inlets can override this setting.
                       (default 0)
        max_buffered -- Optionally the maximum amount of data to buffer (in
                        seconds if there is a nominal sampling rate, otherwise
                        x100 in samples). The default is 6 minutes of data.
                        Note that, for high-bandwidth data, you will want to
                        use a lower value here to avoid running out of RAM.
                        (default 360)

        """
        self.obj = lib.lsl_create_outlet(info.obj, chunk_size, max_buffered)
        self.obj = c_void_p(self.obj)
        if not self.obj:
            raise RuntimeError("could not create stream outlet.")
        self.channel_format = info.channel_format()
        self.channel_count = info.channel_count()
        self.do_push_sample = fmt2push_sample[self.channel_format]
        self.do_push_chunk = fmt2push_chunk[self.channel_format]
        self.value_type = fmt2type[self.channel_format]
        self.sample_type = self.value_type * self.channel_count

    def __del__(self):
        """Destroy an outlet.

        The outlet will no longer be discoverable after destruction and all
        connected inlets will stop delivering data.

        """
        # noinspection PyBroadException
        try:
            lib.lsl_destroy_outlet(self.obj)
        except:
            pass

    def push_sample(self, x, timestamp=0.0, pushthrough=True):
        """Push a sample into the outlet.

        Each entry in the list corresponds to one channel.

        Keyword arguments:
        x -- A list of values to push (one per channel).
        timestamp -- Optionally the capture time of the sample, in agreement
                     with local_clock(); if omitted, the current
                     time is used. (default 0.0)
        pushthrough -- Whether to push the sample through to the receivers
                       instead of buffering it with subsequent samples.
                       Note that the chunk_size, if specified at outlet
                       construction, takes precedence over the pushthrough flag.
                       (default True)

        """
        if len(x) == self.channel_count:
            if self.channel_format == cf_string:
                x = [v.encode('utf-8') for v in x]
            handle_error(self.do_push_sample(self.obj, self.sample_type(*x),
                                             c_double(timestamp),
                                             c_int(pushthrough)))
        else:
            raise ValueError("length of the sample (" + str(len(x)) + ") must "
                             "correspond to the stream's channel count ("
                             + str(self.channel_count) + ").")

    def push_chunk(self, x, timestamp=0.0, pushthrough=True):
        """Push a list of samples into the outlet.

        samples -- A list of samples, either as a list of lists or a list of
                   multiplexed values.
        timestamp -- Optionally the capture time of the most recent sample, in
                     agreement with local_clock(); if omitted, the current
                     time is used. The time stamps of other samples are
                     automatically derived according to the sampling rate of
                     the stream. (default 0.0)
        pushthrough Whether to push the chunk through to the receivers instead
                    of buffering it with subsequent samples. Note that the
                    chunk_size, if specified at outlet construction, takes
                    precedence over the pushthrough flag. (default True)

        """
        try:
            n_values = self.channel_count * len(x)
            data_buff = (self.value_type * n_values).from_buffer(x)
            handle_error(self.do_push_chunk(self.obj, data_buff,
                                            c_long(n_values),
                                            c_double(timestamp),
                                            c_int(pushthrough)))
        except TypeError:
            if len(x):
                if type(x[0]) is list:
                    x = [v for sample in x for v in sample]
                if self.channel_format == cf_string:
                    x = [v.encode('utf-8') for v in x]
                if len(x) % self.channel_count == 0:
                    constructor = self.value_type * len(x)
                    # noinspection PyCallingNonCallable
                    handle_error(self.do_push_chunk(self.obj, constructor(*x),
                                                    c_long(len(x)),
                                                    c_double(timestamp),
                                                    c_int(pushthrough)))
                else:
                    raise ValueError("Each sample must have the same number of channels ("
                                     + str(self.channel_count) + ").")

    def have_consumers(self):
        """Check whether consumers are currently registered.

        While it does not hurt, there is technically no reason to push samples
        if there is no consumer.

        """
        return bool(lib.lsl_have_consumers(self.obj))

    def wait_for_consumers(self, timeout):
        """Wait until some consumer shows up (without wasting resources).

        Returns True if the wait was successful, False if the timeout expired.

        """
        return bool(lib.lsl_wait_for_consumers(self.obj, c_double(timeout)))

    def get_info(self):
        outlet_info = lib.lsl_get_info(self.obj)
        return StreamInfo(handle=outlet_info)


# =========================
# === Resolve Functions ===
# =========================

def resolve_streams(wait_time=1.0):
    """Resolve all streams on the network.

    This function returns all currently available streams from any outlet on
    the network. The network is usually the subnet specified at the local
    router, but may also include a group of machines visible to each other via
    multicast packets (given that the network supports it), or list of
    hostnames. These details may optionally be customized by the experimenter
    in a configuration file (see Network Connectivity in the LSL wiki).

    Keyword arguments:
    wait_time -- The waiting time for the operation, in seconds, to search for
                 streams. Warning: If this is too short (<0.5s) only a subset
                 (or none) of the outlets that are present on the network may
                 be returned. (default 1.0)

    Returns a list of StreamInfo objects (with empty desc field), any of which
    can subsequently be used to open an inlet. The full description can be
    retrieved from the inlet.

    """
    # noinspection PyCallingNonCallable
    buffer = (c_void_p * 1024)()
    num_found = lib.lsl_resolve_all(byref(buffer), 1024, c_double(wait_time))
    return [StreamInfo(handle=buffer[k]) for k in range(num_found)]


def resolve_byprop(prop, value, minimum=1, timeout=FOREVER):
    """Resolve all streams with a specific value for a given property.

    If the goal is to resolve a specific stream, this method is preferred over
    resolving all streams and then selecting the desired one.

    Keyword arguments:
    prop -- The StreamInfo property that should have a specific value (e.g.,
            "name", "type", "source_id", or "desc/manufaturer").
    value -- The string value that the property should have (e.g., "EEG" as
             the type property).
    minimum -- Return at least this many streams. (default 1)
    timeout -- Optionally a timeout of the operation, in seconds. If the
               timeout expires, less than the desired number of streams
               (possibly none) will be returned. (default FOREVER)

    Returns a list of matching StreamInfo objects (with empty desc field), any
    of which can subsequently be used to open an inlet.

    Example: results = resolve_Stream_byprop("type","EEG")

    """
    # noinspection PyCallingNonCallable
    buffer = (c_void_p * 1024)()
    num_found = lib.lsl_resolve_byprop(byref(buffer), 1024,
                                       c_char_p(str.encode(prop)),
                                       c_char_p(str.encode(value)),
                                       minimum,
                                       c_double(timeout))
    return [StreamInfo(handle=buffer[k]) for k in range(num_found)]


def resolve_bypred(predicate, minimum=1, timeout=FOREVER):
    """Resolve all streams that match a given predicate.

    Advanced query that allows to impose more conditions on the retrieved
    streams; the given string is an XPath 1.0 predicate for the <description>
    node (omitting the surrounding []'s), see also
    http://en.wikipedia.org/w/index.php?title=XPath_1.0&oldid=474981951.

    Keyword arguments:
    predicate -- The predicate string, e.g. "name='BioSemi'" or
                "type='EEG' and starts-with(name,'BioSemi') and
                 count(description/desc/channels/channel)=32"
    minimum -- Return at least this many streams. (default 1)
    timeout -- Optionally a timeout of the operation, in seconds. If the
               timeout expires, less than the desired number of streams
               (possibly none) will be returned. (default FOREVER)

    Returns a list of matching StreamInfo objects (with empty desc field), any
    of which can subsequently be used to open an inlet.

    """
    # noinspection PyCallingNonCallable
    buffer = (c_void_p * 1024)()
    num_found = lib.lsl_resolve_bypred(byref(buffer), 1024,
                                       c_char_p(str.encode(predicate)),
                                       minimum,
                                       c_double(timeout))
    return [StreamInfo(handle=buffer[k]) for k in range(num_found)]


# ====================
# === Memory functions
# ====================
def free_char_p_array_memory(char_p_array, num_elements):
    pointers = cast(char_p_array, POINTER(c_void_p))
    for p in range(num_elements):
        if pointers[p] is not None:  # only free initialized pointers
            lib.lsl_destroy_string(pointers[p])


# ====================
# === Stream Inlet ===
# ====================

class StreamInlet:
    """A stream inlet.

    Inlets are used to receive streaming data (and meta-data) from the lab
    network.

    """

    def __init__(self, info, max_buflen=360, max_chunklen=0, recover=True, processing_flags=0):
        """Construct a new stream inlet from a resolved stream description.

        Keyword arguments:
        description -- A resolved stream description object (as coming from one
                of the resolver functions). Note: the stream_inlet may also be
                constructed with a fully-specified stream_info, if the desired
                channel format and count is already known up-front, but this is
                strongly discouraged and should only ever be done if there is
                no time to resolve the stream up-front (e.g., due to
                limitations in the client program).
        max_buflen -- Optionally the maximum amount of data to buffer (in
                      seconds if there is a nominal sampling rate, otherwise
                      x100 in samples). Recording applications want to use a
                      fairly large buffer size here, while real-time
                      applications would only buffer as much as they need to
                      perform their next calculation. (default 360)
        max_chunklen -- Optionally the maximum size, in samples, at which
                        chunks are transmitted (the default corresponds to the
                        chunk sizes used by the sender). Recording programs
                        can use a generous size here (leaving it to the network
                        how to pack things), while real-time applications may
                        want a finer (perhaps 1-sample) granularity. If left
                        unspecified (=0), the sender determines the chunk
                        granularity. (default 0)
        recover -- Try to silently recover lost streams that are recoverable
                   (=those that that have a source_id set). In all other cases
                   (recover is False or the stream is not recoverable)
                   functions may throw a lost_error if the stream's source is
                   lost (e.g., due to an app or computer crash). (default True)
        processing_flags -- Post-processing options. Use one of the post-processing
                   flags `proc_none`, `proc_clocksync`, `proc_dejitter`, `proc_monotonize`,
                   or `proc_threadsafe`. Can also be a logical OR combination of multiple
                   flags. Use `proc_ALL` for all flags. (default proc_none).
        """
        if type(info) is list:
            raise TypeError("description needs to be of type StreamInfo, "
                            "got a list.")
        self.obj = lib.lsl_create_inlet(info.obj, max_buflen, max_chunklen,
                                        recover)
        self.obj = c_void_p(self.obj)
        if not self.obj:
            raise RuntimeError("could not create stream inlet.")
        if processing_flags > 0:
            handle_error(lib.lsl_set_postprocessing(self.obj, processing_flags))
        self.channel_format = info.channel_format()
        self.channel_count = info.channel_count()
        self.do_pull_sample = fmt2pull_sample[self.channel_format]
        self.do_pull_chunk = fmt2pull_chunk[self.channel_format]
        self.value_type = fmt2type[self.channel_format]
        self.sample_type = self.value_type * self.channel_count
        self.sample = self.sample_type()
        self.buffers = {}

    def __del__(self):
        """Destructor. The inlet will automatically disconnect if destroyed."""
        # noinspection PyBroadException
        try:
            lib.lsl_destroy_inlet(self.obj)
        except:
            pass

    def info(self, timeout=FOREVER):
        """Retrieve the complete information of the given stream.

        This includes the extended description. Can be invoked at any time of
        the stream's lifetime.

        Keyword arguments:
        timeout -- Timeout of the operation. (default FOREVER)

        Throws a TimeoutError (if the timeout expires), or LostError (if the
        stream source has been lost).

        """
        errcode = c_int()
        result = lib.lsl_get_fullinfo(self.obj, c_double(timeout),
                                      byref(errcode))
        handle_error(errcode)
        return StreamInfo(handle=result)

    def open_stream(self, timeout=FOREVER):
        """Subscribe to the data stream.

        All samples pushed in at the other end from this moment onwards will be
        queued and eventually be delivered in response to pull_sample() or
        pull_chunk() calls. Pulling a sample without some preceding open_stream
        is permitted (the stream will then be opened implicitly).

        Keyword arguments:
        timeout -- Optional timeout of the operation (default FOREVER).

        Throws a TimeoutError (if the timeout expires), or LostError (if the
        stream source has been lost).

        """
        errcode = c_int()
        lib.lsl_open_stream(self.obj, c_double(timeout), byref(errcode))
        handle_error(errcode)

    def close_stream(self):
        """Drop the current data stream.

        All samples that are still buffered or in flight will be dropped and
        transmission and buffering of data for this inlet will be stopped. If
        an application stops being interested in data from a source
        (temporarily or not) but keeps the outlet alive, it should call
        lsl_close_stream() to not waste unnecessary system and network
        resources.

        """
        lib.lsl_close_stream(self.obj)

    def time_correction(self, timeout=FOREVER):
        """Retrieve an estimated time correction offset for the given stream.

        The first call to this function takes several miliseconds until a
        reliable first estimate is obtained. Subsequent calls are instantaneous
        (and rely on periodic background updates). The precision of these
        estimates should be below 1 ms (empirically within +/-0.2 ms).

        Keyword arguments:
        timeout -- Timeout to acquire the first time-correction estimate
                   (default FOREVER).

        Returns the current time correction estimate. This is the number that
        needs to be added to a time stamp that was remotely generated via
        local_clock() to map it into the local clock domain of this
        machine.

        Throws a TimeoutError (if the timeout expires), or LostError (if the
        stream source has been lost).

        """
        errcode = c_int()
        result = lib.lsl_time_correction(self.obj, c_double(timeout),
                                         byref(errcode))
        handle_error(errcode)
        return result

    def pull_sample(self, timeout=FOREVER, sample=None):
        """Pull a sample from the inlet and return it.

        Keyword arguments:
        timeout -- The timeout for this operation, if any. (default FOREVER)
                   If this is passed as 0.0, then the function returns only a
                   sample if one is buffered for immediate pickup.

        Returns a tuple (sample,timestamp) where sample is a list of channel
        values and timestamp is the capture time of the sample on the remote
        machine, or (None,None) if no new sample was available. To remap this
        time stamp to the local clock, add the value returned by
        .time_correction() to it.

        Throws a LostError if the stream source has been lost. Note that, if
        the timeout expires, no TimeoutError is thrown (because this case is
        not considered an error).

        """

        # support for the legacy API
        if type(timeout) is list:
            assign_to = timeout
            timeout = sample if type(sample) is float else 0.0
        else:
            assign_to = None

        errcode = c_int()
        timestamp = self.do_pull_sample(self.obj, byref(self.sample),
                                        self.channel_count, c_double(timeout),
                                        byref(errcode))
        handle_error(errcode)
        if timestamp:
            sample = [v for v in self.sample]
            if self.channel_format == cf_string:
                sample = [v.decode('utf-8') for v in sample]
            if assign_to is not None:
                assign_to[:] = sample
            return sample, timestamp
        else:
            return None, None

    def pull_chunk(self, timeout=0.0, max_samples=1024, dest_obj=None):
        """Pull a chunk of samples from the inlet.

        Keyword arguments:
        timeout -- The timeout of the operation; if passed as 0.0, then only
                   samples available for immediate pickup will be returned.
                   (default 0.0)
        max_samples -- Maximum number of samples to return. (default
                       1024)
        dest_obj -- A Python object that supports the buffer interface.
                    If this is provided then the dest_obj will be updated in place
                    and the samples list returned by this method will be empty.
                    It is up to the caller to trim the buffer to the appropriate
                    number of samples.
                    A numpy buffer must be order='C'
                    (default None)

        Returns a tuple (samples,timestamps) where samples is a list of samples
        (each itself a list of values), and timestamps is a list of time-stamps.

        Throws a LostError if the stream source has been lost.

        """
        # look up a pre-allocated buffer of appropriate length
        num_channels = self.channel_count
        max_values = max_samples * num_channels

        if max_samples not in self.buffers:
            # noinspection PyCallingNonCallable
            self.buffers[max_samples] = ((self.value_type * max_values)(),
                                         (c_double * max_samples)())
        if dest_obj is not None:
            data_buff = (self.value_type * max_values).from_buffer(dest_obj)
        else:
            data_buff = self.buffers[max_samples][0]
        ts_buff = self.buffers[max_samples][1]

        # read data into it
        errcode = c_int()
        # noinspection PyCallingNonCallable
        num_elements = self.do_pull_chunk(self.obj, byref(data_buff),
                                          byref(ts_buff), max_values,
                                          max_samples, c_double(timeout),
                                          byref(errcode))
        handle_error(errcode)
        # return results (note: could offer a more efficient format in the
        # future, e.g., a numpy array)
        num_samples = num_elements / num_channels
        if dest_obj is None:
            samples = [[data_buff[s * num_channels + c] for c in range(num_channels)]
                       for s in range(int(num_samples))]
            if self.channel_format == cf_string:
                samples = [[v.decode('utf-8') for v in s] for s in samples]
                free_char_p_array_memory(data_buff, max_values)
        else:
            samples = None
        timestamps = [ts_buff[s] for s in range(int(num_samples))]
        return samples, timestamps

    def samples_available(self):
        """Query whether samples are currently available for immediate pickup.

        Note that it is not a good idea to use samples_available() to determine
        whether a pull_*() call would block: to be sure, set the pull timeout
        to 0.0 or an acceptably low value. If the underlying implementation
        supports it, the value will be the number of samples available
        (otherwise it will be 1 or 0).

        """
        return lib.lsl_samples_available(self.obj)

    def flush(self):
        """
        Drop all queued not-yet pulled samples.
        :return: The number of dropped samples.
        """
        return lib.lsl_inlet_flush(self.obj)

    def was_clock_reset(self):
        """Query whether the clock was potentially reset since the last call.

        This is rarely-used function is only needed for applications that
        combine multiple time_correction values to estimate precise clock
        drift if they should tolerate cases where the source machine was
        hot-swapped or restarted.

        """
        return bool(lib.lsl_was_clock_reset(self.obj))


# ===================
# === XML Element ===
# ===================

class XMLElement:
    """A lightweight XML element tree modeling the .desc() field of StreamInfo.

    Has a name and can have multiple named children or have text content as
    value; attributes are omitted. Insider note: The interface is modeled after
    a subset of pugixml's node type and is compatible with it. See also
    http://pugixml.googlecode.com/svn/tags/latest/docs/manual/access.html for
    additional documentation.

    """

    def __init__(self, handle):
        """Construct new XML element from existing handle."""
        self.e = c_void_p(handle)

    # === Tree Navigation ===

    def first_child(self):
        """Get the first child of the element."""
        return XMLElement(lib.lsl_first_child(self.e))

    def last_child(self):
        """Get the last child of the element."""
        return XMLElement(lib.lsl_last_child(self.e))

    def child(self, name):
        """Get a child with a specified name."""
        return XMLElement(lib.lsl_child(self.e, str.encode(name)))

    def next_sibling(self, name=None):
        """Get the next sibling in the children list of the parent node.

        If a name is provided, the next sibling with the given name is returned.

        """
        if name is None:
            return XMLElement(lib.lsl_next_sibling(self.e))
        else:
            return XMLElement(lib.lsl_next_sibling_n(self.e, str.encode(name)))

    def previous_sibling(self, name=None):
        """Get the previous sibling in the children list of the parent node.

        If a name is provided, the previous sibling with the given name is
        returned.

        """
        if name is None:
            return XMLElement(lib.lsl_previous_sibling(self.e))
        else:
            return XMLElement(lib.lsl_previous_sibling_n(self.e,
                                                         str.encode(name)))

    def parent(self):
        """Get the parent node."""
        return XMLElement(lib.lsl_parent(self.e))

    # === Content Queries ===

    def empty(self):
        """Whether this node is empty."""
        return bool(lib.lsl_empty(self.e))

    def is_text(self):
        """Whether this is a text body (instead of an XML element).

        True both for plain char data and CData.

        """
        return bool(lib.lsl_is_text(self.e))

    def name(self):
        """Name of the element."""
        return lib.lsl_name(self.e).decode('utf-8')

    def value(self):
        """Value of the element."""
        return lib.lsl_value(self.e).decode('utf-8')

    def child_value(self, name=None):
        """Get child value (value of the first child that is text).

        If a name is provided, then the value of the first child with the
        given name is returned.

        """
        if name is None:
            res = lib.lsl_child_value(self.e)
        else:
            res = lib.lsl_child_value_n(self.e, str.encode(name))
        return res.decode('utf-8')

    # === Modification ===

    def append_child_value(self, name, value):
        """Append a child node with a given name, which has a (nameless)
        plain-text child with the given text value."""
        return XMLElement(lib.lsl_append_child_value(self.e,
                                                     str.encode(name),
                                                     str.encode(value)))

    def prepend_child_value(self, name, value):
        """Prepend a child node with a given name, which has a (nameless)
        plain-text child with the given text value."""
        return XMLElement(lib.lsl_prepend_child_value(self.e,
                                                      str.encode(name),
                                                      str.encode(value)))

    def set_child_value(self, name, value):
        """Set the text value of the (nameless) plain-text child of a named
        child node."""
        return XMLElement(lib.lsl_set_child_value(self.e,
                                                  str.encode(name),
                                                  str.encode(value)))

    def set_name(self, name):
        """Set the element's name. Returns False if the node is empty."""
        return bool(lib.lsl_set_name(self.e, str.encode(name)))

    def set_value(self, value):
        """Set the element's value. Returns False if the node is empty."""
        return bool(lib.lsl_set_value(self.e, str.encode(value)))

    def append_child(self, name):
        """Append a child element with the specified name."""
        return XMLElement(lib.lsl_append_child(self.e, str.encode(name)))

    def prepend_child(self, name):
        """Prepend a child element with the specified name."""
        return XMLElement(lib.lsl_prepend_child(self.e, str.encode(name)))

    def append_copy(self, elem):
        """Append a copy of the specified element as a child."""
        return XMLElement(lib.lsl_append_copy(self.e, elem.e))

    def prepend_copy(self, elem):
        """Prepend a copy of the specified element as a child."""
        return XMLElement(lib.lsl_prepend_copy(self.e, elem.e))

    def remove_child(self, rhs):
        """Remove a given child element, specified by name or as element."""
        if type(rhs) is XMLElement:
            lib.lsl_remove_child(self.e, rhs.e)
        else:
            lib.lsl_remove_child_n(self.e, rhs)


# ==========================
# === ContinuousResolver ===
# ==========================

class ContinuousResolver:
    """A convenience class resolving streams continuously in the background.

    This object can be queried at any time for the set of streams that are
    currently visible on the network.

    """

    def __init__(self, prop=None, value=None, pred=None, forget_after=5.0):
        """Construct a new continuous_resolver.

        Keyword arguments:
        forget_after -- When a stream is no longer visible on the network
                        (e.g., because it was shut down), this is the time in
                        seconds after which it is no longer reported by the
                        resolver.

        """
        if pred is not None:
            if prop is not None or value is not None:
                raise ValueError("you can only either pass the prop/value "
                                 "argument or the pred argument, but not "
                                 "both.")
            self.obj = lib.lsl_create_continuous_resolver_bypred(str.encode(pred),
                                                                 c_double(forget_after))
        elif prop is not None and value is not None:
            self.obj = lib.lsl_create_continuous_resolver_byprop(str.encode(prop),
                                                                 str.encode(value),
                                                                 c_double(forget_after))
        elif prop is not None or value is not None:
            raise ValueError("if prop is specified, then value must be "
                             "specified, too, and vice versa.")
        else:
            self.obj = lib.lsl_create_continuous_resolver(c_double(forget_after))
        self.obj = c_void_p(self.obj)
        if not self.obj:
            raise RuntimeError("could not create continuous resolver.")

    def __del__(self):
        """Destructor for the continuous resolver."""
        # noinspection PyBroadException
        try:
            lib.lsl_destroy_continuous_resolver(self.obj)
        except:
            pass

    def results(self):
        """Obtain the set of currently present streams on the network.

        Returns a list of matching StreamInfo objects (with empty desc
        field), any of which can subsequently be used to open an inlet.

        """
        # noinspection PyCallingNonCallable
        buffer = (c_void_p * 1024)()
        num_found = lib.lsl_resolver_results(self.obj, byref(buffer), 1024)
        return [StreamInfo(handle=buffer[k]) for k in range(num_found)]


# =========================
# === Error Definitions ===
# =========================

# noinspection PyShadowingBuiltins
class TimeoutError(RuntimeError):
    # note: although this overrides the name of a built-in exception,
    #       this API is retained here for compatiblity with the Python 2.x
    #       version of pylsl
    pass


class LostError(RuntimeError):
    pass


class InvalidArgumentError(RuntimeError):
    pass


class InternalError(RuntimeError):
    pass


def handle_error(errcode):
    """Error handler function. Translates an error code into an exception."""
    if type(errcode) is c_int:
        errcode = errcode.value
    if errcode == 0:
        pass  # no error
    elif errcode == -1:
        raise TimeoutError("the operation failed due to a timeout.")
    elif errcode == -2:
        raise LostError("the stream has been lost.")
    elif errcode == -3:
        raise InvalidArgumentError("an argument was incorrectly specified.")
    elif errcode == -4:
        raise InternalError("an internal error has occurred.")
    elif errcode < 0:
        raise RuntimeError("an unknown error has occurred.")


# =================================================
# === Compatibility Interface for old pylsl API ===
# =================================================

# set class aliases
stream_info = StreamInfo
stream_outlet = StreamOutlet
stream_inlet = StreamInlet
xml_element = XMLElement
timeout_error = TimeoutError
lost_error = LostError
vectorf = vectord = vectorl = vectori = vectors = vectorc = vectorstr = list


def resolve_stream(*args):
    if len(args) == 0:
        return resolve_streams()
    elif type(args[0]) in [int, float]:
        return resolve_streams(args[0])
    elif type(args[0]) is str:
        if len(args) == 1:
            return resolve_bypred(args[0])
        elif type(args[1]) in [int, float]:
            return resolve_bypred(args[0], args[1])
        else:
            if len(args) == 2:
                return resolve_byprop(args[0], args[1])
            else:
                return resolve_byprop(args[0], args[1], args[2])


# ==================================
# === Module Initialization Code ===
# ==================================

def find_liblsl_libraries():
    """Finds the binary lsl library.

    Search order is to first try to use the path stored in the environment
    variable PYLSL_LIB (if available), then search through the package
    directory, and finally search the whole system.

    Returns
    -------
    path: Generator[str]
        A generator yielding possible paths to the library
    """
    if "PYLSL_LIB" in os.environ:
        path = os.environ["PYLSL_LIB"]
        if os.path.isfile(path):
            yield path

    os_name = platform.system()
    if os_name in ["Windows", "Microsoft"]:
        libsuffix = ".dll"
    elif os_name == "Darwin":
        libsuffix = ".dylib"
    elif os_name == "Linux":
        libsuffix = ".so"
    else:
        raise RuntimeError("unrecognized operating system:", os_name)

    # Package-wide search
    libbasepath = os.path.join(os.path.dirname(__file__), "lib")
    for file in os.listdir(libbasepath):
        path = os.path.join(libbasepath, file)
        if os.path.isfile(path) and file.split('.')[-1] == libsuffix[1:]:
            yield path

    # Sytem-wide search, according to docs:
    # On Linux, find_library tries to run external
    # programs (/sbin/ldconfig, gcc, and objdump) to find
    # the library file
    # On OS X, find_library tries several predefined
    # naming schemes and paths to locate the library,
    # On Windows, find_library searches along the system
    # search path. However, we disallow finding system-level
    # lsl.dll on Windows because it causes too many problems
    # and should never be necessary.
    if os_name not in ["Windows", "Microsoft"]:
        for libprefix in ["", "lib"]:
            for debugsuffix in ["", "-debug"]:
                for bitness in ["", str(8 * struct.calcsize("P"))]:
                    path = util.find_library(
                        libprefix + "lsl" + bitness + debugsuffix
                    )
                    if path is not None:
                        yield path


libpaths = find_liblsl_libraries()
while True:
    try:
        libpath = next(libpaths)
        lib = CDLL(libpath)
        break
    except StopIteration:
        raise RuntimeError(
            "Your system configuration is not supported by default by BSL. "
            "Please install 'pylsl' first and make sure it can be imported "
            "on your system.")
    except Exception:
        continue


# set function return types where necessary
lib.lsl_local_clock.restype = c_double
lib.lsl_create_streaminfo.restype = c_void_p
lib.lsl_library_info.restype = c_char_p
lib.lsl_get_name.restype = c_char_p
lib.lsl_get_type.restype = c_char_p
lib.lsl_get_nominal_srate.restype = c_double
lib.lsl_get_source_id.restype = c_char_p
lib.lsl_get_created_at.restype = c_double
lib.lsl_get_uid.restype = c_char_p
lib.lsl_get_session_id.restype = c_char_p
lib.lsl_get_hostname.restype = c_char_p
lib.lsl_get_desc.restype = c_void_p
lib.lsl_get_xml.restype = c_char_p
lib.lsl_create_outlet.restype = c_void_p
lib.lsl_create_inlet.restype = c_void_p
lib.lsl_get_fullinfo.restype = c_void_p
lib.lsl_get_info.restype = c_void_p
lib.lsl_open_stream.restype = c_void_p
lib.lsl_time_correction.restype = c_double
lib.lsl_pull_sample_f.restype = c_double
lib.lsl_pull_sample_d.restype = c_double
lib.lsl_pull_sample_l.restype = c_double
lib.lsl_pull_sample_i.restype = c_double
lib.lsl_pull_sample_s.restype = c_double
lib.lsl_pull_sample_c.restype = c_double
lib.lsl_pull_sample_str.restype = c_double
lib.lsl_pull_sample_buf.restype = c_double
lib.lsl_first_child.restype = c_void_p
lib.lsl_first_child.argtypes = [c_void_p, ]
lib.lsl_last_child.restype = c_void_p
lib.lsl_last_child.argtypes = [c_void_p, ]
lib.lsl_next_sibling.restype = c_void_p
lib.lsl_next_sibling.argtypes = [c_void_p, ]
lib.lsl_previous_sibling.restype = c_void_p
lib.lsl_previous_sibling.argtypes = [c_void_p, ]
lib.lsl_parent.restype = c_void_p
lib.lsl_parent.argtypes = [c_void_p, ]
lib.lsl_child.restype = c_void_p
lib.lsl_child.argtypes = [c_void_p, c_char_p]
lib.lsl_next_sibling_n.restype = c_void_p
lib.lsl_next_sibling_n.argtypes = [c_void_p, c_char_p]
lib.lsl_previous_sibling_n.restype = c_void_p
lib.lsl_previous_sibling_n.argtypes = [c_void_p, c_char_p]
lib.lsl_name.restype = c_char_p
lib.lsl_name.argtypes = [c_void_p, ]
lib.lsl_value.restype = c_char_p
lib.lsl_value.argtypes = [c_void_p, ]
lib.lsl_child_value.restype = c_char_p
lib.lsl_child_value.argtypes = [c_void_p, ]
lib.lsl_child_value_n.restype = c_char_p
lib.lsl_child_value_n.argtypes = [c_void_p, c_char_p]
lib.lsl_append_child_value.restype = c_void_p
lib.lsl_append_child_value.argtypes = [c_void_p, c_char_p, c_char_p]
lib.lsl_prepend_child_value.restype = c_void_p
lib.lsl_prepend_child_value.argtypes = [c_void_p, c_char_p, c_char_p]
# Return type for lsl_set_child_value, lsl_set_name, lsl_set_value is int
lib.lsl_set_child_value.argtypes = [c_void_p, c_char_p, c_char_p]
lib.lsl_set_name.argtypes = [c_void_p, c_char_p]
lib.lsl_set_value.argtypes = [c_void_p, c_char_p]
lib.lsl_append_child.restype = c_void_p
lib.lsl_append_child.argtypes = [c_void_p, c_char_p]
lib.lsl_prepend_child.restype = c_void_p
lib.lsl_prepend_child.argtypes = [c_void_p, c_char_p]
lib.lsl_append_copy.restype = c_void_p
lib.lsl_append_copy.argtypes = [c_void_p, c_void_p]
lib.lsl_prepend_copy.restype = c_void_p
lib.lsl_prepend_copy.argtypes = [c_void_p, c_void_p]
lib.lsl_remove_child_n.argtypes = [c_void_p, c_char_p]
lib.lsl_remove_child.argtypes = [c_void_p, c_void_p]
lib.lsl_destroy_string.argtypes = [c_void_p]
# noinspection PyBroadException
try:
    lib.lsl_pull_chunk_f.restype = c_long
    lib.lsl_pull_chunk_d.restype = c_long
    lib.lsl_pull_chunk_l.restype = c_long
    lib.lsl_pull_chunk_i.restype = c_long
    lib.lsl_pull_chunk_s.restype = c_long
    lib.lsl_pull_chunk_c.restype = c_long
    lib.lsl_pull_chunk_str.restype = c_long
    lib.lsl_pull_chunk_buf.restype = c_long
except:
    print("pylsl: chunk transfer functions not available in your liblsl "
          "version.")
# noinspection PyBroadException
try:
    lib.lsl_create_continuous_resolver.restype = c_void_p
    lib.lsl_create_continuous_resolver_bypred.restype = c_void_p
    lib.lsl_create_continuous_resolver_byprop.restype = c_void_p
except:
    print("pylsl: ContinuousResolver not (fully) available in your liblsl "
          "version.")


# int64 support on windows and 32bit OSes isn't there yet
if struct.calcsize("P") != 4 and platform.system() != 'Windows':
    push_sample_int64 = lib.lsl_push_sample_ltp
    pull_sample_int64 = lib.lsl_pull_sample_l
    push_chunk_int64 = lib.lsl_push_chunk_ltp
    pull_chunk_int64 = lib.lsl_pull_chunk_l
else:
    def push_sample_int64(*_):
        raise NotImplementedError('int64 support isn\'t enabled on your platform')
    pull_sample_int64 = push_chunk_int64 = pull_chunk_int64 = push_sample_int64

# set up some type maps
string2fmt = {'float32': cf_float32, 'double64': cf_double64,
              'string': cf_string, 'int32': cf_int32, 'int16': cf_int16,
              'int8': cf_int8, 'int64': cf_int64}
fmt2string = ['undefined', 'float32', 'double64', 'string', 'int32', 'int16',
              'int8', 'int64']
fmt2type = [[], c_float, c_double, c_char_p, c_int, c_short, c_byte, c_longlong]
fmt2push_sample = [[], lib.lsl_push_sample_ftp, lib.lsl_push_sample_dtp,
                   lib.lsl_push_sample_strtp, lib.lsl_push_sample_itp,
                   lib.lsl_push_sample_stp, lib.lsl_push_sample_ctp, push_sample_int64]
fmt2pull_sample = [[], lib.lsl_pull_sample_f, lib.lsl_pull_sample_d,
                   lib.lsl_pull_sample_str, lib.lsl_pull_sample_i,
                   lib.lsl_pull_sample_s, lib.lsl_pull_sample_c, pull_sample_int64]
# noinspection PyBroadException
try:
    fmt2push_chunk = [[], lib.lsl_push_chunk_ftp, lib.lsl_push_chunk_dtp,
                      lib.lsl_push_chunk_strtp, lib.lsl_push_chunk_itp,
                      lib.lsl_push_chunk_stp, lib.lsl_push_chunk_ctp, push_chunk_int64]
    fmt2pull_chunk = [[], lib.lsl_pull_chunk_f, lib.lsl_pull_chunk_d,
                      lib.lsl_pull_chunk_str, lib.lsl_pull_chunk_i,
                      lib.lsl_pull_chunk_s, lib.lsl_pull_chunk_c, pull_chunk_int64]
except:
    # if not available
    fmt2push_chunk = [None, None, None, None, None, None, None, None]
    fmt2pull_chunk = [None, None, None, None, None, None, None, None]



"""
Fill function docstrings to avoid redundant docstrings in multiple files.
Inspired from mne: https://mne.tools/stable/index.html
Inspired from mne.utils.docs.py by Eric Larson <larson.eric.d@gmail.com>
"""
import sys


# ------------------------- Documentation dictionary -------------------------
docdict = dict()

# -----------------------------------------------
docdict['stream_name'] = """
stream_name : list | str | None
    Servers' name or list of servers' name to connect to.
    If ``None``, connects to all the available streams."""

# -----------------------------------------------
# Stream Receiver
docdict['receiver_get_stream_name'] = """
stream_name : str | None
    Name of the stream from which data is retrieved.
    Can be set to ``None`` (default) if the StreamReceiver is connected to a
    single stream."""
docdict['receiver_get_return_raw'] = """
return_raw : bool
    By default (``False``), data is returned as a `~numpy.array` of shape
    ``(samples, channels)``. If set to ``True``, the StreamReceiver will
    attempt to return data as a MNE Raw instances."""
docdict['receiver_data'] = """
data : `~numpy.array`
    Data ``(samples, channels)``."""
docdict['receiver_timestamps'] = """
timestamps : `~numpy.array`
    Data's timestamps ``(samples, )``."""
docdict['receiver_get_unit'] = """
Returns a raw data array in the unit streamed by the LSL outlet. For conversion
the corresponding scaling factor must be set for each stream, with e.g. for a
stream in uV to convert to V:

.. code-block:: python

    sr.streams['stream_to_convert'].scaling_factor = 1e-6"""
docdict['receiver_bufsize'] = """
bufsize : int | float
    Buffer's size ``[secs]``. ``MAX_BUF_SIZE`` (def: 1-day) is the maximum
    size. Large buffer may lead to a delay if not pulled frequently."""
docdict['receiver_winsize'] = """
winsize : int | float
    Window's size ``[secs]``. Must be smaller than the buffer's size."""

# Not read by sphinx autodoc
docdict['receiver_streamInfo'] = """
streamInfo : LSL StreamInfo.
    Contain all the info from the LSL stream to connect to."""
docdict['receiver_tslist'] = """
tslist : list
    Data's timestamps (samples, )."""

# -----------------------------------------------
# Stream Recorder
docdict['recorder_record_dir'] = """
record_dir : None | path-like
    Path to the directory where data will be saved. If the directory does not
    exist, it is created. If ``None``, the current working directory is used.
"""
docdict['recorder_fname'] = """
fname : None | str
    File name stem used to create the files. The StreamRecorder creates 2 files
    plus an optional third if a software trigger was used, respecting the
    following naming::

      PCL: '{fname}-[stream_name]-raw.pcl'
      FIF: '{fname}-[stream_name]-raw.fif'
      (optional) SOFTWARE trigger events: '{fname}-eve.txt'
"""
docdict['recorder_fif_subdir'] = """
fif_subdir : bool
    If ``True``, the ``.pcl`` files are converted to ``.fif`` in a
    subdirectory ``'fif': record_dir/fif/...`` instead of ``record_dir``."""
docdict['recorder_verbose'] = """
verbose : bool
    If ``True``, a timer showing since when the recorder started is displayed
    every seconds."""

# -----------------------------------------------
# Stream Viewer

# Not read by sphinx autodoc
docdict['viewer_scope'] = """
scope : Scope
    Scope connected to a StreamReceiver acquiring the data and applying
    filtering. The scope has a buffer of _BUFFER_DURATION seconds
    (default: 30s)."""
docdict['viewer_backend_geometry'] = """
geometry : tuple | list
    Window geometry as (pos_x, pos_y, size_x, size_y)."""
docdict['viewer_backend_xRange'] = """
xRange : int
    Range of the x-axis (plotting time duration) in seconds."""
docdict['viewer_backend_yRange'] = """
yRange : float
    Range of the y-axis (amplitude) in uV."""
docdict['viewer_scope_stream_receiver'] = """
stream_receiver : StreamReceiver
    Connected StreamReceiver."""
docdict['viewer_scope_stream_name'] = """
stream_name : str
    Stream to connect to."""
docdict['viewer_event_type'] = """
event_type : str
    Type of event. Supported: 'LPT'."""
docdict['viewer_event_value'] = """
event_value : int
    Value of the event."""
docdict['viewer_position_buffer'] = """
position_buffer : float
    Time (seconds) at which the event is positionned in the buffer where:
        0 represents the older events exiting the buffer.
        _BUFFER_DURATION represents the newer events entering the
        buffer."""
docdict['viewer_position_plot'] = """
position_plot : float
    Time (seconds) at which the event is positionned in the plotting window
    where:
        0 represents the older events exiting the window.
        xRange represents the newer events entering the window."""

# -----------------------------------------------
# Triggers
docdict['trigger_verbose'] = """
verbose : bool
    If ``True``, display a ``logger.info`` message when a trigger is sent."""

# ------------------------- Documentation functions --------------------------
docdict_indented = dict()


def fill_doc(f):
    """
    Fill a docstring with docdict entries.

    Parameters
    ----------
    f : callable
        The function to fill the docstring of. Will be modified in place.

    Returns
    -------
    f : callable
        The function, potentially with an updated ``__doc__``.
    """
    docstring = f.__doc__
    if not docstring:
        return f

    lines = docstring.splitlines()
    indent_count = _indentcount_lines(lines)

    try:
        indented = docdict_indented[indent_count]
    except KeyError:
        indent = ' ' * indent_count
        docdict_indented[indent_count] = indented = dict()

        for name, docstr in docdict.items():
            lines = [indent+line if k != 0 else line
                     for k, line in enumerate(docstr.strip().splitlines())]
            indented[name] = '\n'.join(lines)

    try:
        f.__doc__ = docstring % indented
    except (TypeError, ValueError, KeyError) as exp:
        funcname = f.__name__
        funcname = docstring.split('\n')[0] if funcname is None else funcname
        raise RuntimeError('Error documenting %s:\n%s'
                           % (funcname, str(exp)))

    return f


def _indentcount_lines(lines):
    """
    Minimum indent for all lines in line list.

    >>> lines = [' one', '  two', '   three']
    >>> indentcount_lines(lines)
    1
    >>> lines = []
    >>> indentcount_lines(lines)
    0
    >>> lines = [' one']
    >>> indentcount_lines(lines)
    1
    >>> indentcount_lines(['    '])
    0
    """
    indent = sys.maxsize
    for line in lines:
        line_stripped = line.lstrip()
        if line_stripped:
            indent = min(indent, len(line) - len(line_stripped))
    if indent == sys.maxsize:
        return 0
    return indent


def copy_doc(source):
    """
    Copy the docstring from another function (decorator).

    The docstring of the source function is prepepended to the docstring of the
    function wrapped by this decorator.

    This is useful when inheriting from a class and overloading a method. This
    decorator can be used to copy the docstring of the original method.

    Parameters
    ----------
    source : function
        Function to copy the docstring from

    Returns
    -------
    wrapper : function
        The decorated function

    Examples
    --------
    >>> class A:
    ...     def m1():
    ...         '''Docstring for m1'''
    ...         pass
    >>> class B(A):
    ...     @copy_doc(A.m1)
    ...     def m1():
    ...         ''' this gets appended'''
    ...         pass
    >>> print(B.m1.__doc__)
    Docstring for m1 this gets appended
    """
    def wrapper(func):
        if source.__doc__ is None or len(source.__doc__) == 0:
            raise ValueError('Cannot copy docstring: docstring was empty.')
        doc = source.__doc__
        if func.__doc__ is not None:
            doc += func.__doc__
        func.__doc__ = doc
        return func
    return wrapper



import os
import time
import pickle
import datetime
from pathlib import Path
import multiprocessing as mp

from ..utils import Timer
from ..utils.io import pcl2fif
from ..utils._logs import logger
from ..utils._docs import fill_doc
from ..utils._checks import _check_type
from ..stream_receiver import StreamReceiver, StreamEEG
from ..stream_receiver._stream import MAX_BUF_SIZE


@fill_doc
class StreamRecorder:
    """
    Class for recording the signals coming from LSL streams.

    Parameters
    ----------
    %(recorder_record_dir)s
    %(recorder_fname)s
    %(stream_name)s
    %(recorder_fif_subdir)s
    %(recorder_verbose)s
    """

    def __init__(self, record_dir=None, fname=None, stream_name=None,
                 fif_subdir=True, *, verbose=False):
        self._record_dir = StreamRecorder._check_record_dir(record_dir)
        self._fname = StreamRecorder._check_fname(fname)
        _check_type(stream_name, (None, str, list, tuple),
                    item_name='stream_name')
        self._stream_name = stream_name
        _check_type(fif_subdir, (bool, ), item_name='fif_subdir')
        self._fif_subdir = fif_subdir
        _check_type(verbose, (bool, ), item_name='verbose')
        self._verbose = verbose

        self._eve_file = None  # for SOFTWARE triggers
        self._process = None
        self._state = mp.Value('i', 0)

    def start(self, blocking=True):
        """
        Start the recording in a new process.

        Parameters
        ----------
        blocking : bool
            If ``True``, waits for the child process to start recording data.
        """
        _check_type(blocking, (bool, ), item_name='blocking')
        fname, self._eve_file = StreamRecorder._create_fname(
            self._record_dir, self._fname)
        logger.debug("File name stem is '%s'.", fname)
        logger.debug("Event file name is '%s'.", self._eve_file)

        self._process = mp.Process(
            target=self._record,
            args=(self._record_dir, fname, self._stream_name, self._fif_subdir,
                  self._verbose, self._eve_file, self._state))
        self._process.start()

        if blocking:
            while self._state.value == 0:
                pass

        logger.info(self.__repr__())

    def stop(self):
        """
        Stops the recording.
        """
        if self._process is None:
            logger.warning('StreamRecorder was not started. Skipping.')
            return

        with self._state.get_lock():
            self._state.value = 0

        logger.info('Waiting for StreamRecorder process to finish.')
        self._process.join(10)
        if self._process.is_alive():
            logger.error('Recorder process not finishing..')
            self._process.kill()
            raise RuntimeError
        logger.info('Recording finished.')

        self._eve_file = None
        self._process = None

    def _record(self, record_dir, fname, stream_name, fif_subdir, verbose,
                eve_file, state):
        """
        The function called in the new process.
        Instance a _Recorder and start recording.
        """
        recorder = _Recorder(record_dir, fname, stream_name, fif_subdir,
                             verbose, eve_file, state)
        recorder.record()

    # --------------------------------------------------------------------
    def __enter__(self):
        """Context manager entry point."""
        self.start(blocking=True)

    def __exit__(self, exc_type, exc_value, exc_tracebac):
        """Context manager exit point."""
        self.stop()

    def __repr__(self):
        """Representation of the instance."""
        status = 'ON' if self._state.value == 1 else 'OFF'
        streams = self._stream_name if self._stream_name is not None \
            else 'All streams'
        return f'<Recorder: {streams} | {status} | {self._record_dir}>'

    # --------------------------------------------------------------------
    @staticmethod
    def _check_record_dir(record_dir):
        """
        Converts record_dir to a Path, or select the current working directory
        if record_dir is None.
        """
        _check_type(record_dir, (None, 'path-like'), item_name='record_dir')
        if record_dir is None:
            record_dir = Path.cwd()
        else:
            record_dir = Path(record_dir)
        return record_dir

    @staticmethod
    def _check_fname(fname):
        """
        Checks that the file name stem is a string or None.
        """
        _check_type(fname, (None, str), item_name='fname')
        return fname

    @staticmethod
    def _create_fname(record_dir, fname):
        """
        Creates the file name path using the current datetime if fname is None.
        """
        fname = fname if fname is not None \
            else time.strftime('%Y%m%d-%H%M%S', time.localtime())
        eve_file = record_dir / f'{fname}-eve.txt'

        return fname, eve_file

    # --------------------------------------------------------------------
    @property
    def record_dir(self):
        """
        Path to the directory where data will be saved.

        :type: Path
        """
        return self._record_dir

    @property
    def fname(self):
        """
        File name stem used to create the files. The StreamRecorder creates 2
        files plus an optional third if a software trigger was used, respecting
        the following naming:

        .. code-block:: python

            PCL: '{fname}-[stream_name]-raw.pcl'
            FIF: '{fname}-[stream_name]-raw.fif'
            (optional) SOFTWARE trigger events: '{fname}-eve.txt'

        :type: str
        """
        return self._fname

    @property
    def stream_name(self):
        """
        Servers' name or list of servers' name to connect to.

        :type: str | list
        """
        return self._stream_name

    @property
    def fif_subdir(self):
        """
        If ``True``, the ``.pcl`` files are converted to ``.fif`` in a
        subdirectory ``'fif': record_dir/fif/...`` instead of ``record_dir``.

        :type: bool
        """
        return self._fif_subdir

    @property
    def verbose(self):
        """
        If ``True``, a timer showing since when the recorder started is
        displayed every seconds.

        :type: bool
        """
        return self._verbose

    @property
    def eve_file(self):
        """
        Path to the event file for SoftwareTrigger.

        :type: Path
        """
        return self._eve_file

    @property
    def process(self):
        """
        Launched process.

        :type: Process
        """
        return self._process

    @property
    def state(self):
        """
        Recording state of the recorder

        - ``0``: Not recording.
        - ``1``: Recording.

        :type: `multiprocessing.Value`
        """
        return self._state


@fill_doc
class _Recorder:
    """
    Class creating the .pcl files, recording data through a StreamReceiver and
    saving the data in the .pcl and .fif files.

    Parameters
    ----------
    %(recorder_record_dir)s
    %(recorder_fname)s
    %(stream_name)s
    %(recorder_fif_subdir)s
    %(recorder_verbose)s
    state : mp.Value
        Recording state of the recorder:
            - 0: Not recording.
            - 1: Recording.
        This variable is used to stop the recording from another process.
    eve_file : str | Path
        Path to the event file for SoftwareTrigger.
    """

    def __init__(self, record_dir, fname, stream_name, fif_subdir, verbose,
                 eve_file, state):
        self._record_dir = record_dir
        self._fname = fname
        self._stream_name = stream_name
        self._fif_subdir = fif_subdir
        self._verbose = verbose
        self._eve_file = eve_file
        self._state = state

    def record(self):
        """
        Instantiate a StreamReceiver, create the files, record and save.
        """
        sr = StreamReceiver(
            bufsize=MAX_BUF_SIZE, stream_name=self._stream_name)
        pcl_files = _Recorder._create_files(self._record_dir, self._fname, sr)

        with self._state.get_lock():
            self._state.value = 1

        if self._verbose:
            previous_time = -1  # init to -1 to start log at 00:00:00
            verbose_timer = Timer()

        # Acquisition loop
        while self._state.value == 1:
            sr.acquire()

            if self._verbose:
                if verbose_timer.sec() - previous_time >= 1:
                    previous_time = verbose_timer.sec()
                    duration = str(datetime.timedelta(
                        seconds=int(verbose_timer.sec())))
                    logger.info('RECORDING %s', duration)

        self._save(sr, pcl_files)

    def _save(self, sr, pcl_files):
        """
        Save the data in the StreamReceiver buffer to the .pcl and .fif files.
        """
        logger.info('Saving raw data..')
        for stream in sr.streams:
            signals, timestamps = sr.get_buffer(stream)

            if isinstance(sr.streams[stream], StreamEEG):
                signals[:, 1:] *= 1E-6

            data = {
                'signals': signals,
                'timestamps': timestamps,
                'events': None,
                'sample_rate': sr.streams[stream].sample_rate,
                'channels': len(sr.streams[stream].ch_list),
                'ch_names': sr.streams[stream].ch_list,
                'lsl_time_offset': sr.streams[stream].lsl_time_offset}

            with open(pcl_files[stream], 'wb') as file:
                pickle.dump(data, file, protocol=pickle.HIGHEST_PROTOCOL)

            logger.info("Saved to '%s'", pcl_files[stream])

            if not isinstance(sr.streams[stream], StreamEEG):
                continue
            logger.info('Converting raw files into fif.')

            if self._fif_subdir:
                out_dir = None
            else:
                out_dir = self._record_dir

            if self._eve_file.exists():
                logger.info('Found matching event file, adding events.')
                pcl2fif(
                    pcl_files[stream], out_dir, external_event=self._eve_file)
            else:
                pcl2fif(
                    pcl_files[stream], out_dir, external_event=None)

    # --------------------------------------------------------------------
    @staticmethod
    def _create_files(record_dir, fname, sr):
        """
        Create the .pcl files and check writability.
        """
        os.makedirs(record_dir, exist_ok=True)

        pcl_files = dict()
        for stream in sr.streams:
            pcl_files[stream] = \
                record_dir / f'{fname}-{stream}-raw.pcl'

            try:
                with open(pcl_files[stream], 'w') as file:
                    file.write(
                        'Data will be written when the recording is finished.')
            except Exception as error:
                raise error("Could not write to '%s'. Check permissions"
                            % {pcl_files[stream]})

        logger.info(
            'Record to files: \n' +
            '\n'.join(str(file) for file in pcl_files.values()))

        return pcl_files



"""Utility functions for testing. Inspired from MNE."""

import time
import requests
from pathlib import Path
from functools import partial

import pytest

from ..datasets import (eeg_resting_state, eeg_resting_state_short,
                        eeg_auditory_stimuli, trigger_def)
from ..datasets._fetching import _hashfunc


def requires_good_network(function):
    """Decorator to skip a test if a network connection is not available."""
    try:
        requests.get('https://github.com/', timeout=1)
        skip = False
    except ConnectionError:
        skip = True
    name = function.__name__
    reason = f'Test {name} skipped, requires a good network connection.'
    return pytest.mark.skipif(skip, reason=reason)(function)


def _requires_dataset_or_good_network(function, dataset):
    """Decorator to skip a test if a required dataset is absent and it can not
    be downloaded."""
    # BSL datasets
    try:
        fname = dataset.PATH
        download = False if fname.exists() \
            and _hashfunc(fname, hash_type='md5') == dataset.MD5 else True
    except AttributeError:
        # MNE datasets
        try:
            fname = dataset.data_path(download=False)
            download = False if fname != '' and Path(fname).exists() else True
        except AttributeError:
            raise ValueError('Unsupported dataset.')

    if download:
        return requires_good_network(function)
    else:
        return function


requires_eeg_auditory_stimuli_dataset = partial(
    _requires_dataset_or_good_network, dataset=eeg_auditory_stimuli)
requires_eeg_resting_state_dataset = partial(
    _requires_dataset_or_good_network, dataset=eeg_resting_state)
requires_eeg_resting_state_short_dataset = partial(
    _requires_dataset_or_good_network, dataset=eeg_resting_state_short)
requires_trigger_def_dataset = partial(
    _requires_dataset_or_good_network, dataset=trigger_def)


def requires_parallel(function):
    """Decorator to skip a test if a build-in ParallelPort is not available."""
    try:
        from psychopy.parallel import ParallelPort  # noqa: F401
    except ModuleNotFoundError:
        return pytest.mark.skipif(
            True, reason='psychopy not installed.')(function)
    return pytest.mark.skipif(
        True, reason='ParallelPort tests are not supported.')(function)


def requires_arduino2lpt(function):
    """Decorator to skip a test if an Arduino to LPT converter is not
    available."""
    try:
        import serial
        from serial.tools import list_ports
    except ModuleNotFoundError:
        return pytest.mark.skipif(
            True, reason='pyserial not installed.')(function)
    for arduino in list_ports.grep(regexp='Arduino'):
        try:
            ser = serial.Serial(arduino.device, 115200)
            time.sleep(0.2)
            ser.write(bytes([1]))
            ser.close()
            return function
        except Exception:
            pass
    return pytest.mark.skipif(
        True, reason='Arduino to LPT not found.')(function)



import time

import mne
import pytest

from bsl import StreamRecorder, StreamPlayer, logger, set_log_level
from bsl.datasets import eeg_resting_state
from bsl.triggers import ParallelPortTrigger
from bsl.utils._tests import (requires_eeg_resting_state_dataset,
                              requires_parallel,
                              requires_arduino2lpt)


set_log_level('INFO')
logger.propagate = True

# TODO: Compact the syntax into one parametized function.


@requires_parallel
@requires_eeg_resting_state_dataset
def test_parallel(tmp_path, portaddr):
    """Testing for build-in parallel port."""
    # TODO
    pass


@requires_arduino2lpt
@requires_eeg_resting_state_dataset
def test_parallel_with_arduino(tmp_path, caplog):
    """Testing for Arduino to LPT converter."""
    # Test trigger
    with StreamPlayer('StreamPlayer', eeg_resting_state.data_path()):
        recorder = StreamRecorder(record_dir=tmp_path, fname='test',
                                  stream_name='StreamPlayer', fif_subdir=False)

        trigger = ParallelPortTrigger(address='arduino', verbose=True)
        time.sleep(0.5)
        assert 'using an Arduino converter.' in caplog.text
        assert trigger.verbose
        trigger.verbose = False

        assert trigger.signal(1)
        time.sleep(0.1)
        assert trigger.signal(2)

        trigger.close()
        time.sleep(0.5)
        recorder.stop()

    raw = mne.io.read_raw_fif(tmp_path / 'test-StreamPlayer-raw.fif')
    events = mne.find_events(raw, stim_channel='TRIGGER')
    assert events.shape == (2, 3)
    assert (events[:, 2] == [1, 2]).all()

    # Test delay
    trigger = ParallelPortTrigger(address='arduino', delay=100, verbose=False)
    time.sleep(0.1)
    assert trigger.signal(1)
    assert not trigger.signal(2)
    assert 'new signal before the end of the last' in caplog.text

    # Test property setters
    time.sleep(0.2)
    with pytest.raises(AttributeError, match="can't set attribute"):
        trigger.delay = 50
    assert trigger.delay == 100



import threading
import time

from ._trigger import _Trigger
from ..utils._docs import fill_doc, copy_doc
from ..utils._checks import _check_type
from ..utils._imports import import_optional_dependency
from ..utils._logs import logger


@fill_doc
class ParallelPortTrigger(_Trigger):
    """
    Trigger using a parallel port (also called LPT port).

    Parameters
    ----------
    address : int (hex) | str
        The address of the parallel port on the system. If ``'arduino'``, uses
        an Arduino to LPT converter. Design of the converter can be found here
        :ref:`arduino2lpt`.
    delay : int
        Delay in milliseconds until which a new trigger cannot be sent. During
        this time, the pins remains in the same state.
    %(trigger_verbose)s

    Notes
    -----
    The address is specific to the system. Typical addresses are:

    - On Linux::

          LPT1 = /dev/parport0
          LPT2 = /dev/parport1
          LPT3 = /dev/parport2

    - On Windows, commom port addresses::

          LPT1 = 0x0378 or 0x03BC
          LPT2 = 0x0278 or 0x0378
          LPT3 = 0x0278

    - macOS does not have support for parallel ports.
    """

    def __init__(self, address, delay: int = 50, *, verbose: bool = True):
        _check_type(address, ('int', str), item_name='address')
        _check_type(delay, ('int', ), item_name='delay')
        super().__init__(verbose)

        self._address = address
        self._delay = delay / 1000.0

        # Initialize port
        if self._address == 'arduino':
            self._connect_arduino()
        else:
            self._connect_pport()

        self._offtimer = threading.Timer(self._delay, self._signal_off)

    def _connect_arduino(self, baud_rate: int = 115200):
        """
        Connect to the Arduino LPT converter.
        """
        # Imports
        import_optional_dependency(
            "serial", extra="Install 'pyserial' for ARDUINO support.")

        from serial import Serial, SerialException
        from serial.tools import list_ports

        # Look for arduino
        logger.info('ParallelPort trigger is using an Arduino converter.')
        com_port = None
        for arduino in list_ports.grep(regexp='Arduino'):
            logger.info(f"Found '{arduino}'.")
            com_port = arduino.device
            break
        if com_port is None:
            raise IOError('No arduino card was found.')

        # Connect to arduino
        try:
            self._port = Serial(com_port, baud_rate)
        except SerialException as error:
            logger.error(
                "Disconnect and reconnect the ARDUINO convertor because "
                f"{error}", exc_info=True)
            raise Exception from error

        time.sleep(1)
        logger.info(f'Connected to {com_port}.')

    def _connect_pport(self):
        """
        Connect to the ParallelPort.
        """
        # Imports
        import_optional_dependency(
            "psychopy",
            extra="Install 'psychopy' for Parallel Port support.")

        from psychopy.parallel import ParallelPort

        # Connect to ParallelPort
        logger.info('ParallelPort trigger is using an on-board port.')
        try:
            self._port = ParallelPort(self._address)
        except PermissionError as error:
            logger.error(
                'To fix a PermissionError, try adding your user into the '
                'group with access to the port or try changing the chmod on '
                'the port.')
            raise Exception from error

        time.sleep(1)
        logger.info(f'Connected to {self._address}.')

    @copy_doc(_Trigger.signal)
    def signal(self, value: int) -> bool:
        _check_type(value, ('int', ), item_name='value')
        if self._offtimer.is_alive():
            logger.warning(
                'You are sending a new signal before the end of the last '
                'signal. Signal ignored. Delay required = {self.delay} ms.')
            return False
        self._set_data(value)
        super().signal(value)
        self._offtimer.start()
        return True

    def _signal_off(self):
        """
        Reset trigger signal to 0 and reset offtimer as Threads are one-call
        only.
        """
        self._set_data(0)
        self._offtimer = threading.Timer(self._delay, self._signal_off)

    @copy_doc(_Trigger._set_data)
    def _set_data(self, value: int):
        super()._set_data(value)
        if self._address == 'arduino':
            self._port.write(bytes([value]))
        else:
            self._port.setData(value)

    def close(self):
        """
        Disconnects the parallel port. This method should free the parallel
        port and let other application or python process use it.
        """
        if self._address == 'arduino':
            try:
                self._port.close()
            except Exception:
                pass
        else:
            if hasattr(self, '_port'):
                del self._port

    def __del__(self):
        self.close()

    # --------------------------------------------------------------------
    @property
    def address(self):
        """
        Port address.

        :type: int | str
        """
        return self._address

    @property
    def delay(self):
        """
        Delay to wait between two :meth:`~ParallelPortTrigger.signal` call in
        milliseconds.

        :type: float
        """
        return self._delay * 1000.0



"""
Convert known file format to FIF.
"""
import os
import pickle
from pathlib import Path

import mne
import numpy as np
from mne.io._read_raw import supported

from . import find_event_channel
from ._logs import logger


mne.set_log_level('ERROR')


# ------------------------- Stream Recorder PCL -------------------------
def pcl2fif(fname, out_dir=None, external_event=None,
            precision='double', replace=False, overwrite=True):
    """
    Convert BSL pickle format to MNE Raw format.

    Parameters
    ----------
    fname : file-like
        Pickle file path to convert to ``.fif`` format.
    out_dir : path-like
        Saving directory. If ``None``, it will be the directory
        ``fname.parent/'fif'``.
    external_event : file-like
        Event file path in text format, following MNE event structure. Each row
        should be: ``index 0 event``.
    precision : str
        Data matrix format. ``[single|double|int|short]``, ``'single'``
        improves backward compatability.
    replace : bool
        If `True`, previous events will be overwritten by the new ones from
        the external events file.
    overwrite : bool
        If ``True``, overwrite the previous file.
    """
    fname = Path(fname)
    if not fname.is_file():
        raise IOError('File %s not found.' % fname)
    if fname.suffix != '.pcl':
        raise IOError("File type %s is not '.pcl'." % fname.suffix)

    if out_dir is not None:
        out_dir = Path(out_dir)
    else:
        out_dir = fname.parent / 'fif'
    os.makedirs(out_dir, exist_ok=True)

    fiffile = out_dir / str(fname.stem + '.fif')

    # Load from file
    with open(fname, 'rb') as file:
        data = pickle.load(file)

    # MNE format
    raw = _format_pcl_to_mne_RawArray(data)

    # Add events from txt file
    if external_event is not None:
        events = _load_events_from_txt(raw.times, external_event,
                                       data["timestamps"][0])
        if 0 < len(events):
            raw.add_events(events, stim_channel='TRIGGER', replace=replace)

    # Save
    raw.save(fiffile, verbose=False, overwrite=overwrite, fmt=precision)
    logger.info("Data saved to: '%s'", fiffile)


def _format_pcl_to_mne_RawArray(data):
    """
    Format the raw data to the MNE RawArray structure.
    Data must be recorded with BSL StreamRecorder.

    Parameters
    ----------
    data : dict
        Data loaded from the .pcl file.

    Returns
    -------
    raw : Raw
        MNE raw structure.
    """
    if isinstance(data['signals'], list):
        signals_raw = np.array(data['signals'][0]).T    # to channels x samples
    else:
        signals_raw = data['signals'].T                 # to channels x samples

    sample_rate = data['sample_rate']

    # Look for channels name
    if 'ch_names' not in data:
        ch_names = [f'CH{x+1}' for x in range(signals_raw.shape[0])]
    else:
        ch_names = data['ch_names']

    # search for the trigger channel
    trig_ch = find_event_channel(signals_raw, ch_names)
    # TODO: patch to be improved for multi-trig channel recording
    if isinstance(trig_ch, list):
        trig_ch = trig_ch[0]

    # move trigger channel to index 0
    if trig_ch is None:
        # Add a event channel to index 0 for consistency.
        logger.warning(
            'Event channel was not found. '
            'Adding a blank event channel to index 0.')
        eventch = np.zeros([1, signals_raw.shape[1]])
        signals = np.concatenate((eventch, signals_raw), axis=0)
        # data['channels'] is not reliable any more
        num_eeg_channels = signals_raw.shape[0]
        trig_ch = 0
        ch_names = ['TRIGGER'] + ch_names

    elif trig_ch == 0:
        signals = signals_raw
        num_eeg_channels = data['channels'] - 1

    else:
        logger.info('Moving event channel %s to 0.', trig_ch)
        signals = np.concatenate((signals_raw[[trig_ch]],
                                  signals_raw[:trig_ch],
                                  signals_raw[trig_ch + 1:]),
                                 axis=0)
        assert signals_raw.shape == signals.shape
        num_eeg_channels = data['channels'] - 1
        ch_names.pop(trig_ch)
        trig_ch = 0
        ch_names.insert(trig_ch, 'TRIGGER')
        logger.info('New channel list:')
        for channel in ch_names:
            logger.info('%s', channel)

    ch_info = ['stim'] + ['eeg'] * num_eeg_channels
    info = mne.create_info(ch_names, sample_rate, ch_info)

    # create Raw object
    raw = mne.io.RawArray(signals, info)

    return raw


def _load_events_from_txt(raw_times, eve_file, offset):
    """
    Load events delivered by the software trigger from the event txt file, and
    convert LSL timestamps to indices.
    """

    ts_min = min(raw_times)
    ts_max = max(raw_times)
    events = []

    with open(eve_file, 'r') as file:
        for line in file:
            data = line.strip().split('\t')
            event_ts = float(data[0]) - offset
            event_value = int(data[2])
            next_index = np.searchsorted(raw_times, event_ts)
            if next_index >= len(raw_times):
                logger.warning(
                    'Event %d at time %.3f is out of time range (%.3f - %.3f).'
                    % (event_value, event_ts, ts_min, ts_max))
            else:
                events.append([next_index, 0, event_value])

    return np.array(events)


def _add_events_from_txt(raw, events_index, stim_channel='TRIGGER',
                         replace=False):
    """
    Merge the events extracted from a .txt file to the trigger channel.

    Parameters
    ----------
    raw : Raw
        MNE Raw instance.
    events_index : np.array
        MNE-compatible events [shape=(n_events, 3)].
        Used as input to raw.add_events.
    stim_channel : str
        Stim channel where the events are added.
    replace : bool
        If True, the old events on the stim channel are removed before
        adding the new ones.
    """
    if len(events_index) == 0:
        logger.warning('No events were found in the event file.')
    else:
        logger.info('Found %i events', len(events_index))
        raw.add_events(events_index, stim_channel=stim_channel,
                       replace=replace)


# ------------------------- General converter -------------------------
# Edit readers with BSL '.pcl' reader.
supported['.pcl'] = pcl2fif


def any2fif(fname, out_dir=None, overwrite=True, precision='double'):
    """
    Generic file format converter to Raw.
    Uses `mne.io.read_raw`.

    Parameters
    ----------
    fname : file-like
        File path to convert to ``.fif`` format.
    out_dir : path-like
        Saving directory. If ``None``, it will be the directory
        ``fname.parent/'fif'.``
    overwrite : bool
        If ``True``, overwrite previously converted files with the same name.
    precision : str
        Data matrix format. ``[single|double|int|short]``, ``'single'``
        improves backward compatability.
    """
    fname = Path(fname)
    if not fname.is_file():
        raise IOError('File %s not found.' % fname)
    if fname.suffix not in supported:
        raise IOError('File type %s is not supported.' % fname.suffix)

    if fname.suffix == '.pcl':
        eve_file = fname.parent / (fname.stem[:-4] + 'eve.txt')
        if eve_file.exists():
            logger.info("Adding events from '%s'", eve_file)
        else:
            logger.info("No SOFTWARE event file '%s'", eve_file)
            eve_file = None

        pcl2fif(fname, out_dir=out_dir, external_event=eve_file,
                precision=precision, replace=False, overwrite=overwrite)

    else:
        if out_dir is not None:
            out_dir = Path(out_dir)
        else:
            out_dir = fname.parent / 'fif'
        os.makedirs(out_dir, exist_ok=True)

        fiffile = out_dir / fname.stem + '.fif'

        raw = mne.io.read_raw(fname)
        raw.save(fiffile, verbose=False, overwrite=overwrite, fmt=precision)



from threading import Thread

import mne
import numpy as np

from ._stream import StreamEEG, StreamMarker
from ..externals import pylsl
from ..utils import Timer
from ..utils._logs import logger
from ..utils._docs import fill_doc
from ..utils._checks import _check_type


@fill_doc
class StreamReceiver:
    """
    Class for data acquisition from LSL streams.

    It supports the streams of:
        - EEG
        - Markers

    Parameters
    ----------
    %(receiver_bufsize)s
    %(receiver_winsize)s
    %(stream_name)s
    """

    def __init__(self, bufsize=1, winsize=1, stream_name=None):
        self._connected = False
        self._streams = dict()
        self._acquisition_threads = dict()
        self._mne_infos = dict()

        self._winsize = StreamReceiver._check_winsize(winsize)
        self._bufsize = StreamReceiver._check_bufsize(bufsize, self._winsize)
        self.connect(stream_name)

    @fill_doc
    def connect(self, stream_name=None, timeout=5, force=False):
        """
        Search for the available streams on the LSL network and connect to the
        appropriate ones. If a LSL stream fullfills the requirements (name...),
        a connection is established.

        This function is called while instanciating a StreamReceiver and can be
        recall to reconnect to the LSL streams.

        Parameters
        ----------
        %(stream_name)s
        timeout : int | float
            Timeout duration in seconds after which the search is abandonned.
        force : bool
            If ``True``, force reconnect if the StreamReceiver was already
            connected.
        """
        _check_type(timeout, ('numeric', ), item_name='timeout')
        _check_type(force, (bool, ), item_name='force')

        if not force and self._connected:
            return True

        stream_name = StreamReceiver._check_format_stream_name(stream_name)
        if stream_name is not None and force and self._connected:
            self._stream_name.extend(stream_name)
        else:
            self._stream_name = stream_name

        self._connected = False
        self._streams = dict()

        if self._stream_name is None:
            logger.info(
                "Looking for available LSL streaming servers...")
        else:
            logger.info(
                "Looking for server(s): '%s'...", ', '.join(self._stream_name))

        watchdog = Timer()
        while watchdog.sec() <= timeout:
            streamInfos = pylsl.resolve_streams()
            for streamInfo in streamInfos:

                # connect to a specific amp only?
                if self._stream_name is not None and \
                   streamInfo.name() not in self._stream_name:
                    logger.info('Stream %s skipped.', streamInfo.name())
                    continue

                # EEG stream
                if streamInfo.type().lower() == "eeg":
                    self._streams[streamInfo.name()] = StreamEEG(
                        streamInfo, self._bufsize, self._winsize)
                # Marker stream
                elif streamInfo.nominal_srate() == 0:
                    self._streams[streamInfo.name()] = StreamMarker(
                        streamInfo, self._bufsize, self._winsize)

                self._connected = True

            if self._connected:
                break
        else:
            logger.error(
                'Connection timeout. Could not connect to an LSL stream.')
            return False

        for stream in self._streams:
            if stream not in self._acquisition_threads:
                self._acquisition_threads[stream] = None

            if stream not in self._mne_infos:
                if isinstance(self._streams[stream], StreamEEG):
                    ch_names = self._streams[stream].ch_list
                    sfreq = self._streams[stream].sample_rate
                    ch_types = ['eeg'] * len(ch_names)
                    self._mne_infos[stream] = mne.create_info(
                        ch_names, sfreq, ch_types)
                else:
                    self._mne_infos[stream] = None

        self.show_info()
        logger.info('Ready to receive data from the connected streams.')
        return True

    def show_info(self):
        """
        Display the informations about the connected streams.
        """
        for stream in self._streams:
            logger.info("--------------------------------"
                        "--------------------------------")
            logger.info("The stream %s is connected to:", stream)
            self._streams[stream].show_info()

    def disconnect(self, stream_name=None):
        """
        Disconnects the stream ``stream_name`` from the StreamReceiver.
        If ``stream_name`` is a `list`, disconnects all streams in the list.
        If ``stream_name`` is ``None``, disconnects all streams.

        Parameters
        ----------
        stream_name : str | list | None
            Servers' name or list of servers' name to disconnect from.
            If ``None``, disconnect from all streams.
        """
        if not self._connected:
            return

        stream_name = StreamReceiver._check_format_stream_name(stream_name)
        if stream_name is None:
            stream_name = list(self._streams)

        for stream in list(self._streams):
            if stream not in stream_name:
                continue

            self._streams[stream]._inlet.close_stream()
            del self._streams[stream]
            del self._acquisition_threads[stream]
            if self._stream_name is not None:
                idx = self._stream_name.index(stream)
                del self._stream_name[idx]

        if len(self._streams) == 0:
            self._connected = False

    def acquire(self):
        """
        Read data from the streams and fill their buffer using threading.
        """
        if not self._connected:
            raise RuntimeError('StreamReceiver is not connected to any '
                               'streams.')

        for stream in self._streams:
            if self._acquisition_threads[stream] is not None and \
                    self._acquisition_threads[stream].is_alive():
                continue

            thread = Thread(target=self._streams[stream].acquire, args=[])
            thread.daemon = True
            thread.start()
            self._acquisition_threads[stream] = thread

    @fill_doc
    def get_window(self, stream_name=None, return_raw=False):
        """
        Get the latest window from a stream's buffer.
        If several streams are connected, specify the name.

        Parameters
        ----------
        %(receiver_get_stream_name)s
        %(receiver_get_return_raw)s

        Returns
        -------
        %(receiver_data)s
        %(receiver_timestamps)s

        Notes
        -----
        %(receiver_get_unit)s
        """
        if not self._connected:
            raise RuntimeError('StreamReceiver is not connected to any '
                               'streams.')

        if stream_name is None and len(self._streams) == 1:
            stream_name = list(self._streams)[0]
        elif stream_name is None and len(self._streams) > 1:
            raise RuntimeError(
                'StreamReceiver is connected to multiple streams. Please '
                'provide the stream_name argument.')

        try:
            self._acquisition_threads[stream_name].join()
        except KeyError:
            raise KeyError("StreamReceiver is not connected to '%s'."
                           % stream_name)
        except AttributeError:
            raise AttributeError(
                '.acquire() must be called before .get_window().')

        winsize = self._streams[stream_name].buffer.winsize
        window = np.array(self._streams[stream_name].buffer.data[-winsize:])
        timestamps = np.array(self._streams[
            stream_name].buffer.timestamps[-winsize:])
        if len(timestamps) != winsize:
            logger.warning(
                'The buffer of %s does not contain enough samples. Returning '
                'the available samples.', stream_name)

        if len(timestamps) > 0:
            if bool(return_raw) and self._mne_infos[stream_name] is not None:
                window = mne.io.RawArray(
                    window.T, self._mne_infos[stream_name])
                window._filenames = [f'BSL {stream_name}']
            elif bool(return_raw) and self._mne_infos[stream_name] is None:
                logger.warning(
                    'The LSL stream %s can not be converted to MNE raw '
                    'instance. Returning numpy arrays.', stream_name)
        else:
            logger.warning(
                'The LSL stream %s did not return any data.'
                'Returning empty numpy arrays.', stream_name)
            window = np.empty((0, len(self._streams[stream_name].ch_list)))
            timestamps = np.array([])

        return window, timestamps

    @fill_doc
    def get_buffer(self, stream_name=None, return_raw=False):
        """
        Get the entire buffer of a stream.
        If several streams are connected, specify the name.

        Parameters
        ----------
        %(receiver_get_stream_name)s
        %(receiver_get_return_raw)s

        Returns
        -------
        %(receiver_data)s
        %(receiver_timestamps)s

        Notes
        -----
        %(receiver_get_unit)s
        """
        if not self._connected:
            raise RuntimeError('StreamReceiver is not connected to any '
                               'streams.')

        if stream_name is None and len(self._streams) == 1:
            stream_name = list(self._streams)[0]
        elif stream_name is None and len(self._streams) > 1:
            raise RuntimeError(
                'StreamReceiver is connected to multiple streams. Please '
                'provide the stream_name argument.')

        try:
            self._acquisition_threads[stream_name].join()
        except KeyError:
            raise KeyError("StreamReceiver is not connected to '%s'."
                           % stream_name)
        except AttributeError:
            raise AttributeError(
                '.acquire() must be called before .get_buffer().')

        window = np.array(self._streams[stream_name].buffer.data)
        timestamps = np.array(self._streams[stream_name].buffer.timestamps)
        if len(self._streams[stream_name].buffer.timestamps) > 0:
            if bool(return_raw) and self._mne_infos[stream_name] is not None:
                window = mne.io.RawArray(
                    window.T, self._mne_infos[stream_name])
                window._filenames = [f'BSL {stream_name}']
            elif bool(return_raw) and self._mne_infos[stream_name] is None:
                logger.warning(
                    'The LSL stream %s can not be converted to MNE raw '
                    'instance. Returning numpy arrays.', stream_name)
        else:
            logger.warning(
                'The LSL stream %s did not return any data.'
                'Returning empty numpy arrays.', stream_name)
            window = np.empty((0, len(self._streams[stream_name].ch_list)))
            timestamps = np.array([])

        return window, timestamps

    def _get_buffer(self):
        """
        Get the entire buffer of the only connected stream.
        This method is intended for use by the StreamViewer.
        """
        stream_name = list(self._streams)[0]
        self._acquisition_threads[stream_name].join()
        window = np.array(self._streams[stream_name].buffer.data)
        timestamps = np.array(self._streams[stream_name].buffer.timestamps)
        if len(self._streams[stream_name].buffer.timestamps) == 0:
            window = np.empty((0, len(self._streams[stream_name].ch_list)))
            timestamps = np.array([])
        return window, timestamps

    def reset_buffer(self, stream_name=None):
        """
        Clear the stream's buffer.

        Parameters
        ----------
        stream_name : str | list | None
            Name of the stream(s) to reset its buffer.
            If ``None``, reset all stream's buffer.
        """
        stream_name = StreamReceiver._check_format_stream_name(stream_name)
        if stream_name is None:
            stream_name = list(self._streams)

        for stream in self._streams:
            if stream not in stream_name:
                continue

            self._streams[stream].buffer.reset_buffer()

    # --------------------------------------------------------------------
    def __del__(self):
        """Destructor method."""
        self.disconnect()

    def __repr__(self):
        """Representation of the instance."""
        status = 'ON' if self._connected else 'OFF'
        if self._connected:
            streams = str(tuple(self._streams))
        else:
            streams = '()' if self._stream_name is None \
                else str(tuple(self._stream_name))
        repr_str = f'<Receiver: {streams} | {status} | ' + \
                   f'buf: {self._bufsize}s - win: {self._winsize}s>'
        return repr_str

    # --------------------------------------------------------------------
    @staticmethod
    def _check_bufsize(bufsize, winsize):
        """
        Check that bufsize is positive and bigger than the winsize.
        """
        _check_type(bufsize, ('numeric', ), item_name='bufsize')
        bufsize = float(bufsize)
        if bufsize <= 0:
            raise ValueError(
                'Argument bufsize must be a strictly positive int or a '
                'float. Provided: %s' % bufsize)
        if bufsize < winsize:
            logger.error(
                'Buffer size %.2f is smaller than window size. '
                'Setting to %.2f.', bufsize, winsize)
            bufsize = winsize

        return bufsize

    @staticmethod
    def _check_winsize(winsize):
        """
        Check that winsize is positive.
        """
        _check_type(winsize, ('numeric', ), item_name='winsize')
        winsize = float(winsize)
        if winsize <= 0:
            raise ValueError(
                'Argument winsize must be a strictly positive int or a '
                'float. Provided: %s' % winsize)

        return winsize

    @staticmethod
    def _check_format_stream_name(stream_name):
        """
        Check the format of stream_name.
        """
        _check_type(stream_name, (None, str, list, tuple),
                    item_name='stream_name')
        if isinstance(stream_name, (list, tuple)):
            stream_name = list(stream_name)
            if not all(isinstance(name, str) for name in stream_name):
                raise TypeError(
                    'Argument stream_name must be a string or a list of '
                    'strings. Provided: %s' % stream_name)
        elif isinstance(stream_name, str):
            stream_name = [stream_name]

        return stream_name

    # --------------------------------------------------------------------
    @property
    def winsize(self):
        """
        Window's size ``[sec]``.

        :type: int | float
        """
        return self._winsize

    @property
    def bufsize(self):
        """
        Buffer's size ``[sec]``.

        :type: int | float
        """
        return self._bufsize

    @property
    def stream_name(self):
        """
        Connected stream's name.

        :type: None | list
        """
        return self._stream_name

    @property
    def connected(self):
        """
        Connected status.

        :type: bool
        """
        return self._connected

    @property
    def mne_infos(self):
        """
        Dictionary containing the Info for the compatible streams.

        :type: dict
        """
        return self._mne_infos

    @property
    def streams(self):
        """
        Connected streams dictionary ``{stream_name: _Stream}``.

        :type: dict
        """
        return self._streams



import os
from pathlib import Path
from configparser import ConfigParser

from ..utils._logs import logger
from ..utils._checks import _check_type


class TriggerDef:
    """
    Class used to store pairs {str: int} of events name and events value.
    Each name and each value is unique. The pairs can be read from a ``.ini``
    file or edited manually with :meth:`TriggerDef.add` and
    :meth:`TriggerDef.remove`.

    The class will expose the name as attributes ``self.event_str = event_int``
    for all pairs.

    Parameters
    ----------
    trigger_file : None | path-like
        Path to the ``.ini`` file containing the table converting event numbers
        into event strings.

        .. note:: The ``.ini`` file is read with `configparser` and has to be
                  structured as follows:

                  .. code-block:: python

                      [events]
                      event_str_1 = event_id_1   # comment
                      event_str_2 = event_id_2   # comment

                  Example:

                  .. code-block:: python

                      [events]
                      rest = 1
                      stim = 2
    """

    def __init__(self, trigger_file=None):
        self._by_name = dict()
        self._by_value = dict()
        self.read(trigger_file)

    def read(self, trigger_file):
        """
        Read events from a ``.ini`` trigger definition file.

        .. note:: The ``.ini`` file is read with `configparser` and has to be
                  structured as follows:

                  .. code-block:: python

                      [events]
                      event_str_1 = event_id_1   # comment
                      event_str_2 = event_id_2   # comment

                  Example:

                  .. code-block:: python

                      [events]
                      rest = 1
                      stim = 2

        Parameters
        ----------
        trigger_file : path-like
            Path to the ``.ini`` file containing the table converting event
            numbers into event strings.
        """
        self._trigger_file = TriggerDef._check_trigger_file(trigger_file)
        if self._trigger_file is None:
            return

        config = ConfigParser(inline_comment_prefixes=('#', ';'))
        config.optionxform = str
        config.read(str(self._trigger_file))

        for name, value in config.items('events'):
            value = int(value)
            if value in self._by_value:
                logger.info('Event value %s already exists. Skipping.', value)
                continue
            setattr(self, name, value)
            self._by_name[name] = value
            self._by_value[value] = name

    def write(self, trigger_file):
        """
        Write events to a ``.ini`` trigger definition file.

        .. note:: The ``.ini`` file is writen with `configparser` and is
                  structured as follows:

                  .. code-block:: python

                      [events]
                      event_str_1 = event_id_1
                      event_str_2 = event_id_2
        """
        trigger_file = TriggerDef._check_write_to_trigger_file(trigger_file)

        config = ConfigParser()
        config['events'] = self._by_name
        with open(trigger_file, 'w') as configfile:
            config.write(configfile)

    def add(self, name, value, overwrite=False):
        """
        Add an event to the trigger definition instance.

        Parameters
        ----------
        name : str
            Name of the event
        value : int
            Value of the event
        overwrite : bool
            If ``True``, overwrite previous event with the same name or value.
        """
        _check_type(name, (str, ), item_name='name')
        _check_type(value, ('int', ), item_name='value')
        _check_type(overwrite, (bool, ), item_name='overwrite')
        if name in self._by_name and not overwrite:
            logger.info('Event name %s already exists. Skipping.', name)
            return
        if value in self._by_value and not overwrite:
            logger.info('Event value %s already exists. Skipping.', value)
            return

        if name in self._by_name:
            self.remove(name)
        if value in self._by_value:
            self.remove(value)

        setattr(self, name, value)
        self._by_name[name] = value
        self._by_value[value] = name

    def remove(self, event):
        """
        Remove an event from the trigger definition instance.
        The event can be given by name (str) or by value (int).

        Parameters
        ----------
        event : str | int
            If a str is provided, assumes event is the name.
            If a int is provided, assumes event is the value.
        """
        _check_type(event, (str, 'numeric'), item_name='event')
        if isinstance(event, str):
            if event not in self._by_name:
                logger.info('Event name %s not found.', event)
                return
            value = self._by_name[event]
            delattr(self, event)
            del self._by_name[event]
            del self._by_value[value]
        elif isinstance(event, (int, float)):
            event = int(event)
            if event not in self._by_value:
                logger.info('Event value %s not found.', event)
                return
            name = self._by_value[event]
            delattr(self, name)
            del self._by_name[name]
            del self._by_value[event]

    def __repr__(self):
        """Representation of the stored events."""
        if len(self._by_name) == 0:
            return 'TriggerDef: No event found.'
        repr_ = f'TriggerDef: {len(self._by_name)} events.\n'
        for name, value in self._by_name.items():
            repr_ += f'  {name}: {value}\n'
        return repr_

    # --------------------------------------------------------------------
    @staticmethod
    def _check_trigger_file(trigger_file):
        """
        Checks that the provided file exists and ends with .ini. Else returns
        None.
        """
        _check_type(trigger_file, (None, 'path-like'),
                    item_name='trigger_file')

        if trigger_file is None:
            return None
        else:
            trigger_file = Path(trigger_file)

        if trigger_file.exists() and trigger_file.suffix == '.ini':
            logger.info(
                "Found trigger definition file '%s'", trigger_file.name)
            return trigger_file
        elif trigger_file.exists() and trigger_file.suffix != '.ini':
            logger.error(
                'Argument trigger_file must be a valid Path to a .ini file. '
                'Provided: %s', trigger_file.suffix)
            return None
        else:
            logger.error(
                "Trigger event definition file '%s' not found.", trigger_file)
            return None

    @staticmethod
    def _check_write_to_trigger_file(trigger_file):
        """
        Checks that the directory exists and that the file name ends with
        .ini.
        """
        _check_type(trigger_file, ('path-like', ), item_name='trigger_file')

        trigger_file = Path(trigger_file)
        if trigger_file.suffix != '.ini':
            raise ValueError(
                'Argument trigger_file must end with .ini. '
                'Provided: %s' % trigger_file.suffix)

        os.makedirs(trigger_file.parent, exist_ok=True)
        return trigger_file

    # --------------------------------------------------------------------
    @property
    def by_name(self):
        """
        A dictionary with string keys and integers value.

        :type: dict
        """
        return self._by_name

    @property
    def by_value(self):
        """
        A dictionary with integers keys and string values.

        :type: dict
        """
        return self._by_value



import math
import time
from abc import ABC, abstractmethod

import numpy as np

from ._buffer import Buffer
from ..externals import pylsl
from ..utils import Timer
from ..utils._logs import logger
from ..utils import find_event_channel
from ..utils.lsl import lsl_channel_list
from ..utils._checks import _check_type
from ..utils._docs import fill_doc, copy_doc


_MAX_PYLSL_STREAM_BUFSIZE = 10  # max 10 sec of data buffered for LSL
MAX_BUF_SIZE = 86400  # 24h max buffer length
HIGH_LSL_OFFSET_THRESHOLD = 0.1  # Threshold above which the offset is high


@fill_doc
class _Stream(ABC):
    """
    Abstract class representing a base StreamReceiver stream.

    Parameters
    ----------
    %(receiver_streamInfo)s
    %(receiver_bufsize)s
    %(receiver_winsize)s
    """

    @abstractmethod
    def __init__(self, streamInfo, bufsize, winsize):
        winsize = _Stream._check_winsize(winsize)
        bufsize = _Stream._check_bufsize(bufsize, winsize)

        self._streamInfo = streamInfo
        self._sample_rate = self._streamInfo.nominal_srate()
        self._lsl_bufsize = min(_MAX_PYLSL_STREAM_BUFSIZE, bufsize)

        if self._sample_rate is not None:
            samples_per_sec = self._sample_rate
            # max_buflen: seconds
            self._inlet = pylsl.StreamInlet(
                streamInfo,
                max_buflen=math.ceil(self._lsl_bufsize))
        else:
            samples_per_sec = 100
            # max_buflen: samples x100
            self._inlet = pylsl.StreamInlet(
                streamInfo,
                max_buflen=math.ceil(self._lsl_bufsize*samples_per_sec))
        self._inlet.open_stream()

        self._extract_stream_info()
        self._create_ch_name_list()
        self._lsl_time_offset = None

        self._watchdog = Timer()
        self._blocking = True
        self._blocking_time = 5

        # Convert to samples
        bufsize = _Stream._convert_sec_to_samples(bufsize, samples_per_sec)
        winsize = _Stream._convert_sec_to_samples(winsize, samples_per_sec)
        self._lsl_bufsize = _Stream._convert_sec_to_samples(
            self._lsl_bufsize, samples_per_sec)
        self._buffer = Buffer(bufsize, winsize)

    def _create_ch_name_list(self):
        """
        Create the channels' name list.
        """
        self._ch_list = lsl_channel_list(self._inlet)

        if not self._ch_list:
            self._ch_list = [f"ch_{i+1}" for i in range(
                self._streamInfo.channel_count())]

    def _extract_stream_info(self):
        """
        Extract the name, serial number and if it's a slave.
        """
        self._name = self._streamInfo.name()
        self._serial = self._inlet.info().desc().child(
            'acquisition').child_value('serial_number')

        if self._serial == '':
            self._serial = 'N/A'

        self._is_slave = self._inlet.info().desc().child(
            'amplifier').child('settings').child(
                'is_slave').first_child().value() == 'true'

    def show_info(self):
        """
        Display the stream's info.
        """
        logger.info(f'Server: {self._name}({self._serial}) '
                    f'/ type:{self._streamInfo.type()} '
                    f'@ {self._streamInfo.hostname()} '
                    f'(v{self._streamInfo.version()}).')
        logger.info(f'Source sampling rate: {self._sample_rate}')
        logger.info(f'Channels: {self._streamInfo.channel_count()}')

        # Check for high LSL offset
        if self._lsl_time_offset is None:
            logger.info(
                'No LSL timestamp offset computed, no data received yet.')
        elif abs(self._lsl_time_offset) > HIGH_LSL_OFFSET_THRESHOLD:
            logger.warning(
                f'LSL server {self._name}({self._serial}) has a high '
                f'timestamp offset [offset={self._lsl_time_offset:.3f}].')
        else:
            logger.info(
                f'LSL server {self._name}({self._serial}) synchronized '
                f'[offset={self._lsl_time_offset:.3f}]')

    @abstractmethod
    @fill_doc
    def acquire(self):
        """
        Pull data from the stream's inlet and fill the buffer.

        Returns
        -------
        chunk : list
            Data [samples x channels]
        %(receiver_tslist)s
        """
        chunk = []
        tslist = []
        received = False
        self._watchdog.reset()

        # Acquire the data
        while not received:
            while self._watchdog.sec() < self._blocking_time:
                if len(tslist) == 0:
                    chunk, tslist = self._inlet.pull_chunk(
                        timeout=0.0, max_samples=self._lsl_bufsize)
                    if not self._blocking and len(tslist) == 0:
                        received = True
                        break
                if len(tslist) > 0:
                    if self._lsl_time_offset is None:  # First acquisition
                        self._compute_offset(tslist)
                    received = True
                    tslist = self._correct_lsl_offset(tslist)
                    break
                time.sleep(0.0005)  # TODO: test without
            else:
                # Give up and return empty values to avoid deadlock
                logger.error(
                    f'Timeout occurred [{self._blocking_time}secs] while '
                    f'acquiring data from {self._name}({self._serial}). ')
                received = True

        return chunk, tslist

    @fill_doc
    def _compute_offset(self, tslist):
        """
        Compute the LSL offset coming from some devices.

        It has to be called just after acquiring the data/timestamps in order
        to be valid.

        Parameters
        ----------
        %(receiver_tslist)s
        """
        self._lsl_time_offset = tslist[-1] - pylsl.local_clock()
        return self._lsl_time_offset

    @fill_doc
    def _correct_lsl_offset(self, tslist):
        """
        Correct the timestamps if there is a high LSL offset.

        Parameters
        ----------
        %(receiver_tslist)s
        """
        if abs(self._lsl_time_offset) > HIGH_LSL_OFFSET_THRESHOLD:
            tslist = [t - self._lsl_time_offset for t in tslist]

        return tslist

    # --------------------------------------------------------------------
    @staticmethod
    @fill_doc
    def _check_winsize(winsize):
        """
        Check that winsize is positive.

        Parameters
        ----------
        %(receiver_winsize)s
        """
        _check_type(winsize, ('numeric', ), item_name='winsize')
        if winsize <= 0:
            raise ValueError('Invalid window size %s.' % winsize)

        return winsize

    @staticmethod
    @fill_doc
    def _check_bufsize(bufsize, winsize):
        """
        Check that bufsize is positive and bigger than winsize.

        Parameters
        ----------
        %(receiver_bufsize)s
        %(receiver_winsize)s
        """
        _check_type(bufsize, ('numeric', ), item_name='bufsize')

        if bufsize <= 0 or bufsize > MAX_BUF_SIZE:
            logger.error(
                f'Improper buffer size {bufsize:.1f}. '
                f'Setting to {MAX_BUF_SIZE:.1f}.')
            bufsize = MAX_BUF_SIZE
        elif bufsize < winsize:
            logger.error(
                f'Buffer size  {bufsize:.1f} is smaller than window size. '
                f'Setting to {winsize:.1f}.')
            bufsize = winsize

        return bufsize

    @staticmethod
    @fill_doc
    def _convert_sec_to_samples(bufsec, sample_rate):
        """
        Convert a buffer's/window's size from sec to samples.

        Parameters
        ----------
        %(receiver_bufsize)s
        sample_rate : float
             Sampling rate of the LSL server.
             If irregular sampling rate, empirical number of samples per sec.

        Returns
        -------
        int
            Buffer_size's size [samples].
        """
        return round(bufsec * sample_rate)

    # --------------------------------------------------------------------
    @property
    def streamInfo(self):
        """
        Stream info received from the LSL inlet.
        """
        return self._streamInfo

    @property
    def sample_rate(self):
        """
        Stream's sampling rate.
        """
        return self._sample_rate

    @property
    def name(self):
        """
        Stream's name.
        """
        return self._name

    @property
    def serial(self):
        """
        Stream's serial number.
        """
        return self._serial

    @property
    def is_slave(self):
        """
        Value stored in LSL ['amplifier']['settings']['is_slave'].
        """
        return self._is_slave

    @property
    def ch_list(self):
        """
        Channels' name list.
        """
        return self._ch_list

    @property
    def lsl_time_offset(self):
        """
        The difference between the local and the stream's LSL clocks, used for
        timestamps correction.

        Some LSL servers (like OpenVibe) often have a bug of sending its own
        running time instead of LSL time.
        """
        return self._lsl_time_offset

    @property
    def blocking(self):
        """
        If True, the stream wait to receive data.

        :setter: Change the blocking status.
        :type: bool
        """
        return self._blocking

    @blocking.setter
    def blocking(self, blocking):
        self._blocking = bool(blocking)

    @property
    def blocking_time(self):
        """
        If blocking is True, how long to wait to receive data in seconds.

        :setter: Change the blocking time duration (seconds).
        :type: float
        """
        return self._blocking_time

    @blocking_time.setter
    def blocking_time(self, blocking_time):
        self._blocking_time = float(blocking_time)

    @property
    def buffer(self):
        """
        Buffer containing the data and the timestamps.
        """
        return self._buffer


@fill_doc
class StreamMarker(_Stream):
    """
    Class representing a StreamReceiver markers stream.

    Notice the irregular sampling rate.
    This stream is instanciated as non-blocking.

    Parameters
    -----------
    %(receiver_streamInfo)s
    %(receiver_bufsize)s
    %(receiver_winsize)s
    """

    def __init__(self, streamInfo, bufsize=1, winsize=1):
        super().__init__(streamInfo, bufsize, winsize)

        self._blocking = False
        self._blocking_time = float('inf')

    def acquire(self):
        """
        Pull data from the stream's inlet and fill the buffer.
        """
        chunk, tslist = super().acquire()
        self._buffer.fill(chunk, tslist)  # Fill its buffer


@fill_doc
class StreamEEG(_Stream):
    """
    Class representing a StreamReceiver EEG stream.

    This stream is instanciated as blocking.

    Parameters
    -----------
    %(receiver_streamInfo)s
    %(receiver_bufsize)s
    %(receiver_winsize)s
    """

    def __init__(self, streamInfo, bufsize=1, winsize=1):
        super().__init__(streamInfo, bufsize, winsize)

        # self._scaling_factor = 10 ** -6  # change uV -> V unit
        self._scaling_factor = 1

    @copy_doc(_Stream._create_ch_name_list)
    def _create_ch_name_list(self):
        """
         Trigger channel will always move to the first position.
        """
        super()._create_ch_name_list()

        self._find_lsl_trig_ch()
        self._lsl_eeg_channels = list(range(len(self._ch_list)))

        if self._lsl_tr_channel is not None:
            self._ch_list.pop(self._lsl_tr_channel)
            self._lsl_eeg_channels.pop(self._lsl_tr_channel)

        self._ch_list = ['TRIGGER'] + self._ch_list

    def _find_lsl_trig_ch(self):
        """
        Look for the trigger channel index at the LSL inlet level.
        """
        if 'USBamp' in self._name:
            self._lsl_tr_channel = 16

        elif 'BioSemi' in self._name:
            self._lsl_tr_channel = 0

        elif 'SmartBCI' in self._name:
            self._lsl_tr_channel = 23

        elif 'openvibeSignal' in self._name:
            # TODO: Test if this is correct or should be 1E6
            self._scaling_factor = 10E6
            self._lsl_tr_channel = find_event_channel(ch_names=self._ch_list)

        elif 'openvibeMarkers' in self._name:
            self._lsl_tr_channel = find_event_channel(ch_names=self._ch_list)

        elif 'actiCHamp' in self._name:
            self._lsl_tr_channel = -1

        else:
            self._lsl_tr_channel = find_event_channel(ch_names=self._ch_list)

        # TODO: patch to be improved for multi-trig channel recording
        if isinstance(self._lsl_tr_channel, list):
            self._lsl_tr_channel = self._lsl_tr_channel[0]

        if self._lsl_tr_channel is not None:
            logger.debug('Trigger channel idx: %d', self._lsl_tr_channel)
        else:
            logger.debug('Trigger channel was not found.')

    def acquire(self):
        """
        Pull data from the stream's inlet and fill the buffer.
        """
        chunk, tslist = super().acquire()

        if len(chunk) == 0:
            return

        # TODO: Is it not more efficient to keep working on lists?
        data = np.array(chunk)

        # BioSemi has pull-up resistor instead of pull-down
        if 'BioSemi' in self._name and self._lsl_tr_channel is not None:
            datatype = data.dtype
            data[:, self._lsl_tr_channel] = (
                np.bitwise_and(255, data[:, self._lsl_tr_channel].astype(
                    int, copy=False)) - 1).astype(datatype, copy=False)

        # multiply values (to change unit)
        if self._scaling_factor != 1:
            data[:, self._lsl_eeg_channels] *= self._scaling_factor

        if self._lsl_tr_channel is not None:
            # move trigger channel to 0 and add back to the buffer
            data = np.concatenate(
                (data[:, self._lsl_tr_channel].reshape(-1, 1),
                 data[:, self._lsl_eeg_channels]), axis=1)
        else:
            # add an empty channel with zeros to channel 0
            data = np.concatenate((np.zeros((data.shape[0], 1)),
                                   data[:, self._lsl_eeg_channels]), axis=1)

        data = data.tolist()

        # Fill its buffer
        self._buffer.fill(data, tslist)

    # --------------------------------------------------------------------
    @property
    def scaling_factor(self):
        """
        Scaling factor applied to the data to convert to the desired unit.

        :setter: Change the scaling factor applied to the data.
        :type: float
        """
        return self._scaling_factor

    @scaling_factor.setter
    def scaling_factor(self, scaling_factor):
        _check_type(scaling_factor, ('numeric', ), item_name='scaling_factor')
        if scaling_factor <= 0:
            raise ValueError('Property scaling_factor must be a strictly '
                             'positive number. Provided: %s' % scaling_factor)
        self._scaling_factor = scaling_factor



import time

from ._checks import _check_type


class Timer:
    """
    Timer class.

    Parameters
    ----------
    autoreset : bool
        If ``autoreset=True``, timer is reset after any member function call.
    """

    def __init__(self, autoreset=False):
        self.autoreset = autoreset
        self.reset()

    def sec(self):
        """
        Provide the time since reset in seconds.
        """
        read = time.time() - self.ref
        if self._autoreset:
            self.reset()
        return read

    def msec(self):
        """
        Provide the time since reset in milliseconds.
        """
        return self.sec() * 1000.0

    def reset(self):
        """
        Reset the timer to zero.
        """
        self.ref = time.time()

    def sleep_atleast(self, sec):
        """
        Sleep up to sec seconds.
        It's more convenient if ``autoreset=True``.

        Parameters
        ----------
        sec : float
            Time to sleep in seconds.
        """
        timer_sec = self.sec()

        if timer_sec < sec:
            time.sleep(sec - timer_sec)
            if self._autoreset:
                self.reset()

    @property
    def autoreset(self):
        """
        Autoreset status.

        :type: bool
        """
        return self._autoreset

    @autoreset.setter
    def autoreset(self, autoreset):
        _check_type(autoreset, (bool, ), item_name='autoreset')
        self._autoreset = autoreset



"""Utility functions for checking types and values. Inspired from MNE."""

import os
import operator
from pathlib import Path

import numpy as np


def _ensure_int(item, item_name=None):
    """
    Ensure a variable is an integer.

    Parameters
    ----------
    item : object
        Item to check.
    item_name : str | None
        Name of the item to show inside the error message.

    Raises
    ------
    TypeError
        When the type of the item is not int.
    """
    # This is preferred over numbers.Integral, see:
    # https://github.com/scipy/scipy/pull/7351#issuecomment-299713159
    try:
        # someone passing True/False is much more likely to be an error than
        # intentional usage
        if isinstance(item, bool):
            raise TypeError
        item = int(operator.index(item))
    except TypeError:
        item_name = "Item" if item_name is None else "'%s'" % item_name
        raise TypeError("%s must be an int, got %s instead."
                        % (item_name, type(item)))

    return item


class _IntLike:
    @classmethod
    def __instancecheck__(cls, other):
        try:
            _ensure_int(other)
        except TypeError:
            return False
        else:
            return True


class _Callable:
    @classmethod
    def __instancecheck__(cls, other):
        return callable(other)


_types = {
    'numeric': (np.floating, float, _IntLike()),
    'path-like': (str, Path, os.PathLike),
    'int': (_IntLike(), ),
    'callable': (_Callable(), ),
}


def _check_type(item, types, item_name=None):
    """
    Check that item is an instance of types.

    Parameters
    ----------
    item : object
        Item to check.
    types : tuple of types | tuple of str
        Types to be checked against.
        If str, must be one of:
            ('int', 'str', 'numeric', 'path-like', 'callable')
    item_name : str | None
        Name of the item to show inside the error message.

    Raises
    ------
    TypeError
        When the type of the item is not one of the valid options.
    """
    check_types = sum(((type(None), ) if type_ is None else (type_, )
                       if not isinstance(type_, str) else _types[type_]
                       for type_ in types), ())

    if not isinstance(item, check_types):
        type_name = ["None" if cls_ is None else cls_.__name__
                     if not isinstance(cls_, str) else cls_
                     for cls_ in types]
        if len(type_name) == 1:
            type_name = type_name[0]
        elif len(type_name) == 2:
            type_name = ' or '.join(type_name)
        else:
            type_name[-1] = "or " + type_name[-1]
            type_name = ", ".join(type_name)
        item_name = "Item" if item_name is None else "'%s'" % item_name
        raise TypeError(f"{item_name} must be an instance of {type_name}, "
                        f"got {type(item)} instead.")

    return item


def _check_value(item, allowed_values, item_name=None, extra=None):
    """
    Check the value of a parameter against a list of valid options.

    Parameters
    ----------
    item : object
        Item to check.
    allowed_values : tuple of objects
        Allowed values to be checked against.
    item_name : str | None
        Name of the item to show inside the error message.
    extra : str | None
        Extra string to append to the invalid value sentence, e.g.
        "when using ico mode".

    Raises
    ------
    ValueError
        When the value of the item is not one of the valid options.
    """
    if item not in allowed_values:
        item_name = "" if item_name is None else " '%s'" % item_name
        extra = "" if extra is None else " " + extra
        msg = ("Invalid value for the{item_name} parameter{extra}. "
               '{options}, but got {item!r} instead.')
        allowed_values = tuple(allowed_values)  # e.g., if a dict was given
        if len(allowed_values) == 1:
            options = "The only allowed value is %s" % repr(allowed_values[0])
        elif len(allowed_values) == 2:
            options = "Allowed values are %s and %s" % \
                (repr(allowed_values[0]), repr(allowed_values[1]))
        else:
            options = "Allowed values are "
            options += ", ".join([f"{repr(v)}" for v in allowed_values[:-1]])
            options += f", and {repr(allowed_values[-1])}"
        raise ValueError(msg.format(item_name=item_name, extra=extra,
                                    options=options, item=item))

    return item



import numpy as np
from mne.io import BaseRaw
from mne.epochs import BaseEpochs

from ._checks import _check_type


def find_event_channel(inst=None, ch_names=None):
    """
    Find the event channel using heuristics.

    .. warning::

        Not 100% guaranteed to find it.
        If ``inst`` is ``None``, ``ch_names`` must be given.
        If ``inst`` is an MNE instance, ``ch_names`` is ignored if some
        channels types are ``'stim'``.

    Parameters
    ----------
    inst : None | Raw | Epochs | `~numpy.array`
        Data instance. If a `~numpy.array` is provided, the shape must be
        ``(n_channels, n_samples)``.
    ch_names : None | list
        Channels name list.

    Returns
    -------
    event_channel : int | list | None
        Event channel index, list of event channel indexes or ``None`` if not
        found.
    """
    _check_type(inst, (None, np.ndarray, BaseRaw, BaseEpochs),
                item_name='inst')
    _check_type(ch_names, (None, list, tuple), item_name='ch_names')

    # numpy array + ch_names
    if isinstance(inst, np.ndarray) and ch_names is not None:
        tchs = _search_in_ch_names(ch_names)

    # numpy array without ch_names
    elif isinstance(inst, np.ndarray) and ch_names is None:
        # data range between 0 and 255 and all integers?
        tchs = [idx for idx in range(inst.shape[0])
                if (inst[idx].astype(int, copy=False) == inst[idx]).all()
                and max(inst[idx]) <= 255 and min(inst[idx]) == 0]

    # For MNE raw/epochs + ch_names
    elif isinstance(inst, (BaseRaw, BaseEpochs)) and ch_names is not None:
        tchs = [idx for idx, type_ in enumerate(inst.get_channel_types())
                if type_ == 'stim']
        if len(tchs) == 0:
            tchs = _search_in_ch_names(ch_names)

    # For MNE raw/epochs without ch_names
    elif isinstance(inst, (BaseRaw, BaseEpochs)) and ch_names is None:
        tchs = [idx for idx, type_ in enumerate(inst.get_channel_types())
                if type_ == 'stim']
        if len(tchs) == 0:
            tchs = _search_in_ch_names(inst.ch_names)

    # For unknown data type
    elif inst is None:
        if ch_names is None:
            raise ValueError('ch_names cannot be None when inst is None.')
        tchs = _search_in_ch_names(ch_names)

    # output
    if len(tchs) == 0:
        return None
    elif len(tchs) == 1:
        return tchs[0]
    else:
        return tchs


def _search_in_ch_names(ch_names):
    """Search trigger channel by name in a list of valid names."""
    valid_trigger_ch_names = ['TRIGGER', 'STI', 'TRG', 'CH_Event']

    tchs = list()
    for idx, ch_name in enumerate(ch_names):
        if any(trigger_ch_name in ch_name
               for trigger_ch_name in valid_trigger_ch_names):
            tchs.append(idx)

    return tchs



import time
from pathlib import Path
import multiprocessing as mp

import mne
import numpy as np

from ..externals import pylsl
from ..triggers import TriggerDef
from ..utils._logs import logger
from ..utils import find_event_channel
from ..utils._checks import _check_type


class StreamPlayer:
    """
    Class for playing a recorded file on LSL network in another process.

    Parameters
    ----------
    stream_name : str
        Stream's server name, displayed on LSL network.
    fif_file : file-like
        Path to the compatible raw ``.fif`` file to play.
    repeat : int | ``float('inf')``
        Number of times the stream player will loop on the FIF file before
        interrupting. Default ``float('inf')`` can be passed to never interrupt
        streaming.
    trigger_def : None | file-like | TriggerDef
        If not ``None``, a TriggerDef instance is used to log events with a
        descriptive string instead of their ID. If not ``None``, either a
        TriggerDef instance or the path to a valid ``.ini`` file passed to
        TriggerDef.
    chunk_size : int
        Number of samples to send at once (usually ``16-32`` is good enough).
    high_resolution : bool
        If ``True``, it uses `~time.perf_counter` instead of `~time.sleep`
        for higher time resolution. However, it uses more CPU.
    """

    def __init__(self, stream_name, fif_file, repeat=float('inf'),
                 trigger_def=None, chunk_size=16, high_resolution=False):
        _check_type(stream_name, (str, ), item_name='stream_name')
        self._stream_name = stream_name
        self._fif_file = StreamPlayer._check_fif_file(fif_file)
        self._repeat = StreamPlayer._check_repeat(repeat)
        self._trigger_def = StreamPlayer._check_trigger_def(trigger_def)
        self._chunk_size = StreamPlayer._check_chunk_size(chunk_size)
        _check_type(high_resolution, (bool, ), item_name='high_resolution')
        self._high_resolution = high_resolution

        self._process = None
        self._state = mp.Value('i', 0)

    def start(self, blocking=True):
        """
        Start streaming data on LSL network in a new process.

        Parameters
        ----------
        blocking : bool
            If ``True``, waits for the child process to start streaming data.
        """
        _check_type(blocking, (bool, ), item_name='blocking')
        raw = mne.io.read_raw_fif(self._fif_file, preload=True, verbose=False)

        logger.info('Streaming started.')
        self._process = mp.Process(
            target=self._stream,
            args=(self._stream_name, raw, self._repeat,
                  self._trigger_def, self._chunk_size, self._high_resolution,
                  self._state))
        self._process.start()

        if blocking:
            while self._state.value == 0:
                pass

        logger.info(self.__repr__())

    def stop(self):
        """
        Stop the streaming, by terminating the process.
        """
        if self._process is None:
            logger.warning('StreamPlayer was not started. Skipping.')
            return

        with self._state.get_lock():
            self._state.value = 0

        logger.info('Waiting for StreamPlayer %s process to finish.',
                    self._stream_name)
        self._process.join(10)
        if self._process.is_alive():
            logger.error('StreamPlayer process not finishing..')
            self._process.kill()
            raise RuntimeError
        logger.info('Streaming finished.')

        self._process = None

    def _stream(self, stream_name, raw, repeat, trigger_def, chunk_size,
                high_resolution, state):
        """
        The function called in the new process.
        Instance a _Streamer and start streaming.
        """
        streamer = _Streamer(
            stream_name, raw, repeat, trigger_def, chunk_size,
            high_resolution, state)
        streamer.stream()

    # --------------------------------------------------------------------
    def __enter__(self):
        """Context manager entry point."""
        self.start(blocking=True)

    def __exit__(self, exc_type, exc_value, exc_tracebac):
        """Context manager exit point."""
        self.stop()

    def __repr__(self):
        """Representation of the instance."""
        status = 'ON' if self._state.value == 1 else 'OFF'
        return f'<Player: {self._stream_name} | {status} | {self._fif_file}>'

    # --------------------------------------------------------------------
    @staticmethod
    def _check_fif_file(fif_file):
        """
        Check if the provided fif_file is valid.
        """
        _check_type(fif_file, ('path-like', ), item_name='fif_file')
        mne.io.read_raw_fif(fif_file, preload=False, verbose=None)
        return Path(fif_file)

    @staticmethod
    def _check_repeat(repeat):
        """
        Checks that repeat is either infinity or a strictly positive integer.
        """
        if repeat == float('inf'):
            return repeat
        _check_type(repeat, ('int', ), item_name='repeat')
        if repeat <= 0:
            raise ValueError('Argument repeat must be a strictly positive '
                             'integer. Provided: %i' % repeat)
        return repeat

    @staticmethod
    def _check_trigger_def(trigger_def):
        """
        Checks that the trigger file is either a path to a valid trigger
        definition file, in which case it is loader and pass as a TriggerDef,
        or a TriggerDef instance.
        """
        _check_type(trigger_def, (None, TriggerDef, 'path-like'),
                    item_name='trigger_def')
        if isinstance(trigger_def, (type(None), TriggerDef)):
            return trigger_def
        else:
            trigger_def = Path(trigger_def)
            if not trigger_def.exists():
                raise ValueError('Argument trigger_def is a path that does '
                                 'not exist. Provided: %s' % trigger_def)
            trigger_def = TriggerDef(trigger_def)
            return trigger_def

    @staticmethod
    def _check_chunk_size(chunk_size):
        """
        Checks that chunk_size is a strictly positive integer.
        """
        _check_type(chunk_size, ('int', ), item_name='chunk_size')
        if chunk_size <= 0:
            raise ValueError('Argument chunk_size must be a strictly positive '
                             'integer. Provided: %i' % chunk_size)
        if chunk_size not in (16, 32):
            logger.warning(
                'The chunk size %i is different from the usual '
                'values 16 or 32.', chunk_size)
        return chunk_size

    # --------------------------------------------------------------------
    @property
    def stream_name(self):
        """
        Stream's server name, displayed on LSL network.

        :type: str
        """
        return self._stream_name

    @property
    def fif_file(self):
        """
        Path to the compatible raw ``.fif`` file to play.

        :type: str | Path
        """
        return self._fif_file

    @property
    def repeat(self):
        """
        Number of times the stream player will loop on the FIF file before
        interrupting. Default ``float('inf')`` can be passed to never interrupt
        streaming.

        :type: int | ``float('ìnf')``
        """
        return self._repeat

    @property
    def trigger_def(self):
        """
        Either ``None`` or TriggerDef instance converting event numbers into
        event strings.

        :type: TriggerDef
        """
        return self._trigger_def

    @property
    def chunk_size(self):
        """
        Number of samples to send at once (usually ``16-32`` is good enough)
        ``[samples]``.

        :type: int
        """
        return self._chunk_size

    @property
    def high_resolution(self):
        """
        If ``True``, it uses `~time.perf_counter` instead of `~time.sleep`
        for higher time resolution. However, it uses more CPU.

        :type: bool
        """
        return self._high_resolution

    @property
    def process(self):
        """
        Launched streaming process.

        :type: Process
        """
        return self._process

    @property
    def state(self):
        """
        Streaming state of the player:
            - ``0``: Not streaming.
            - ``1``: Streaming.

        :type: `multiprocessing.Value`
        """
        return self._state


class _Streamer:
    """
    Class for playing a recorded file on LSL network.
    """

    def __init__(self, stream_name, raw, repeat, trigger_def,
                 chunk_size, high_resolution, state):
        self._stream_name = stream_name
        self._raw = raw
        self._repeat = repeat
        self._trigger_def = trigger_def
        self._chunk_size = chunk_size
        self._high_resolution = high_resolution
        self._state = state

        self._sinfo = _Streamer._create_lsl_info(
            stream_name=self._stream_name,
            channel_count=len(self._raw.ch_names),
            nominal_srate=self._raw.info['sfreq'],
            ch_names=self._raw.ch_names)
        self._tch = find_event_channel(inst=self._raw)
        # TODO: patch to be improved for multi-trig channel recording
        if isinstance(self._tch, list):
            self._tch = self._tch[0]
        self._scale_raw_data()
        self._outlet = pylsl.StreamOutlet(
            self._sinfo, chunk_size=self._chunk_size)

    def _scale_raw_data(self):
        """
        Assumes raw data is in Volt and convert to microvolts.

        # TODO: Base the scaling on the units in the raw info
        """
        idx = np.arange(self._raw._data.shape[0]) != self._tch
        self._raw._data[idx, :] = self._raw.get_data()[idx, :] * 1E6

    def stream(self):
        """
        Stream data on LSL network.
        """
        idx_chunk = 0
        t_chunk = self._chunk_size / self._raw.info['sfreq']
        finished = False

        if self._high_resolution:
            t_start = time.perf_counter()
        else:
            t_start = time.time()

        played = 0

        with self._state.get_lock():
            self._state.value = 1

        # Streaming loop
        while self._state.value == 1:

            idx_current = idx_chunk * self._chunk_size
            idx_next = idx_current + self._chunk_size
            chunk = self._raw._data[:, idx_current:idx_next]
            data = chunk.transpose().tolist()

            if idx_current >= self._raw._data.shape[1] - self._chunk_size:
                finished = True

            _Streamer._sleep(self._high_resolution, idx_chunk, t_start,
                             t_chunk)

            self._outlet.push_chunk(data)
            logger.debug(
                '[%8.3fs] sent %d samples (LSL %8.3f)'
                % (time.perf_counter(), len(data), pylsl.local_clock()))

            self._log_event(chunk)
            idx_chunk += 1

            if finished:
                idx_chunk = 0
                finished = False
                if self._high_resolution:
                    t_start = time.perf_counter()
                else:
                    t_start = time.time()
                played += 1

                if played < self._repeat:
                    logger.info('Reached the end of data. Restarting.')
                else:
                    logger.info('Reached the end of data. Stopping.')
                    with self._state.get_lock():
                        self._state.value = 0

    def _log_event(self, chunk):
        """
        Look for an event on the data chunk and log it.
        """
        if self._tch is not None:
            event_values = set(chunk[self._tch]) - set([0])

            if len(event_values) > 0:
                if self._trigger_def is None:
                    logger.info('Events: %s', event_values)
                else:
                    for event_value in event_values:
                        if event_value in self._trigger_def.by_value:
                            logger.info(
                                'Events: %s (%s)',
                                event_value,
                                self._trigger_def.by_value[event_value])
                        else:
                            logger.info(
                                'Events: %s (Undefined event)',
                                event_value)

    # --------------------------------------------------------------------
    @staticmethod
    def _create_lsl_info(stream_name, channel_count, nominal_srate, ch_names):
        """
        Extract information from raw and set the LSL server's information
        needed to create the LSL stream.
        """
        sinfo = pylsl.StreamInfo(
            stream_name, channel_count=channel_count, channel_format='float32',
            nominal_srate=nominal_srate, type='EEG', source_id=stream_name)

        desc = sinfo.desc()
        channel_desc = desc.append_child("channels")
        for channel in ch_names:
            channel_desc.append_child('channel')\
                        .append_child_value('label', str(channel))\
                        .append_child_value('type', 'EEG')\
                        .append_child_value('unit', 'microvolts')

        desc.append_child('amplifier')\
            .append_child('settings')\
            .append_child_value('is_slave', 'false')

        desc.append_child('acquisition')\
            .append_child_value('manufacturer', 'BSL')\
            .append_child_value('serial_number', 'N/A')

        return sinfo

    @staticmethod
    def _sleep(high_resolution, idx_chunk, t_start, t_chunk):
        """
        Determine the time to sleep.
        """
        # if a resolution over 2 KHz is needed.
        if high_resolution:
            t_sleep_until = t_start + idx_chunk * t_chunk
            while time.perf_counter() < t_sleep_until:
                pass
        # time.sleep() can have 500 us resolution.
        else:
            t_wait = t_start + idx_chunk * t_chunk - time.time()
            if t_wait > 0.001:
                time.sleep(t_wait)



import numpy as np
from scipy.signal import butter, sosfilt, sosfilt_zi

from ._scope import _Scope
from ...utils._logs import logger
from ...utils._docs import fill_doc, copy_doc
from ...utils import find_event_channel


BP_ORDER = 2


@fill_doc
class ScopeEEG(_Scope):
    """
    Class representing an EEG scope.

    Parameters
    ----------
    %(viewer_scope_stream_receiver)s
    %(viewer_scope_stream_name)s
    """

    # ---------------------------- INIT ----------------------------
    def __init__(self, stream_receiver, stream_name):
        super().__init__(stream_receiver, stream_name)

        # Infos from stream
        tch = find_event_channel(
            ch_names=self._sr.streams[self._stream_name].ch_list)
        # TODO: patch to be improved for multi-trig channel recording
        if isinstance(tch, list):
            tch = tch[0]
        assert tch is not None  # sanity-check

        self._channels_labels = \
            [channel for k, channel in enumerate(
                self._sr.streams[self._stream_name].ch_list) if k != tch]
        self._nb_channels = len(self._channels_labels)

        # Variables
        self._apply_car = False
        self._apply_bandpass = False
        self._selected_channels = list(range(self._nb_channels))

        # Buffers
        self._trigger_buffer = np.zeros(self._duration_buffer_samples)
        self._data_buffer = np.zeros(
            (self._nb_channels, self._duration_buffer_samples),
            dtype=np.float32)

    def init_bandpass_filter(self, low, high):
        """
        Initialize the bandpass filter. The filter is a butter filter of order
        BP_ORDER (default 2).

        Parameters
        ----------
        low : int | float
            Frequency at which the signal is high-passed.
        high : int | float
            Frequency at which the signal is low-passed.
        """
        bp_low = low / (0.5 * self._sample_rate)
        bp_high = high / (0.5 * self._sample_rate)
        self._sos = butter(BP_ORDER, [bp_low, bp_high],
                           btype='band', output='sos')
        self._zi_coeff = sosfilt_zi(
            self._sos).reshape((self._sos.shape[0], 2, 1))
        self._zi = None

    # -------------------------- Main Loop -------------------------
    @copy_doc(_Scope.update_loop)
    def update_loop(self):
        self._read_lsl_stream()
        if len(self._ts_list) > 0:
            self._filter_signal()
            self._filter_trigger()
            # shape (channels, samples)
            self._data_buffer = np.roll(self._data_buffer, -len(self._ts_list),
                                        axis=1)
            self._data_buffer[:, -len(self._ts_list):] = self._data_acquired.T
            # shape (samples, )
            self._trigger_buffer = np.roll(
                self._trigger_buffer, -len(self._ts_list))
            self._trigger_buffer[-len(self._ts_list):] = self._trigger_acquired

    @copy_doc(_Scope._read_lsl_stream)
    def _read_lsl_stream(self):
        """
         The acquired data is splitted between the trigger channel and the data
        channels.
        """
        super()._read_lsl_stream()
        # Remove trigger ch - shapes (samples, ) and (samples, channels)
        self._trigger_acquired = self._data_acquired[:, 0]
        self._data_acquired = self._data_acquired[:, 1:].reshape(
            (-1, self._nb_channels))

    def _filter_signal(self):
        """
        Apply bandpass and CAR filter to the signal acquired if needed.
        """
        if self._apply_bandpass:
            if self._zi is None:
                logger.debug('Initialize ZI coefficient for BP.')
                # Multiply by DC offset
                self._zi = self._zi_coeff*np.mean(self._data_acquired, axis=0)
            self._data_acquired, self._zi = sosfilt(
                self._sos, self._data_acquired, 0, self._zi)

        if self._apply_car and len(self._selected_channels) >= 2:
            car_ch = np.mean(
                self._data_acquired[:, self._selected_channels], axis=1)
            self._data_acquired -= car_ch.reshape((-1, 1))

    def _filter_trigger(self, tol=0.05):
        """
        Cleans up the trigger signal by removing successive duplicates of a
        trigger value.
        """
        self._trigger_acquired[
            np.abs(np.diff(self._trigger_acquired, prepend=[0])) <= tol] = 0

    # --------------------------------------------------------------------
    @property
    def channels_labels(self):
        """
        List of the channel labels present in the connected stream.
        The TRIGGER channel is removed.
        """
        return self._channels_labels

    @property
    def nb_channels(self):
        """
        Number of channels present in the connected stream.
        The TRIGGER channel is removed.
        """
        return self._nb_channels

    @property
    def apply_car(self):
        """
        Boolean. Applies CAR if True.
        """
        return self._apply_car

    @apply_car.setter
    def apply_car(self, apply_car):
        self._apply_car = bool(apply_car)

    @property
    def apply_bandpass(self):
        """
        Boolean. Applies bandpass filter if True.
        """
        return self._apply_bandpass

    @apply_bandpass.setter
    def apply_bandpass(self, apply_bandpass):
        self._apply_bandpass = bool(apply_bandpass)

    @property
    def selected_channels(self):
        """
        List of indices of the selected channels.
        """
        return self._selected_channels

    @selected_channels.setter
    def selected_channels(self, selected_channels):
        self._selected_channels = selected_channels

    @property
    def data_buffer(self):
        """
        Data buffer (channels, samples).
        """
        return self._data_buffer

    @property
    def trigger_buffer(self):
        """
        Trigger buffer (samples, ).
        """
        return self._trigger_buffer



#!/usr/bin/env python

import itertools

from pathlib import Path
from setuptools import setup, find_packages


# Version
version = None
with open(Path(__file__).parent/'bsl'/'_version.py', 'r') as file:
    for line in file:
        line = line.strip()
        if line.startswith('__version__'):
            version = line.split('=')[1].strip().strip('\'')
            break

if version is None:
    raise RuntimeError('Could not determine version.')


# Descriptions
short_description = """BrainStreamingLayer real-time framework for online neuroscience research through LSL-compatible devices."""
long_description_file = Path(__file__).parent / 'README.md'
with open(long_description_file, 'r') as file:
    long_description = file.read()
if long_description_file.suffix == '.md':
    long_description_content_type='text/markdown'
elif long_description_file.suffix == '.rst':
    long_description_content_type='text/x-rst'
else:
    long_description_content_type='text/plain'


# Variables
NAME = 'bsl'
DESCRIPTION = short_description
LONG_DESCRIPTION = long_description
LONG_DESCRIPTION_CONTENT_TYPE=long_description_content_type
AUTHOR = 'Kyuhwa Lee, Arnaud Desvachez, Mathieu Scheltienne'
AUTHOR_EMAIL = 'lee.kyuh@gmail.com, arnaud.desvachez@gmail.com, mathieu.scheltienne@gmail.com'
MAINTAINER = 'Mathieu Scheltienne'
MAINTAINER_EMAIL = 'mathieu.scheltienne@gmail.com'
URL = 'https://github.com/bsl-tools/bsl'
LICENSE = 'LGPL-2.1 - The GNU General Public License'
DOWNLOAD_URL = 'https://github.com/bsl-tools/bsl'
VERSION = version


# Dependencies
def get_requirements(path):
    """Get mandatory dependencies from file."""
    install_requires = list()
    with open(path, 'r') as file:
        for line in file:
            if line.startswith('#'):
                continue
            req = line.strip()
            if len(line) == 0:
                continue
            install_requires.append(req)

    return install_requires


def get_requirements_extras(path, add_all=True):
    """Map extra dependencies to the functionalities."""
    extras_require = dict()
    with open(path, 'r') as file:
        for line in file:
            line = line.strip()
            if line.startswith('#') or len(line) == 0:
                continue
            req, tags = line.split('#')
            for tag in tags.split(','):
                tag = tag.strip()
                if tag not in extras_require:
                    extras_require[tag] = [req.strip()]
                else:
                    extras_require[tag].append(req.strip())

    # add tag 'all' at the end
    if add_all:
        extras_require['all'] = set(itertools.chain(*extras_require.values()))

    return extras_require


install_requires = get_requirements('requirements.txt')
extras_require = get_requirements_extras(
    'requirements_extras.txt', add_all=True)


setup(
    name=NAME,
    version=VERSION,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    maintainer=MAINTAINER,
    maintainer_email=MAINTAINER_EMAIL,
    license=LICENSE,
    url=URL,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type=LONG_DESCRIPTION_CONTENT_TYPE,
    classifiers=[
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Topic :: Software Development',
        'Topic :: Scientific/Engineering',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: Unix',
        'Operating System :: MacOS',
        'Programming Language :: Python :: 3 :: Only',
        'Natural Language :: English'
        ],
    keywords='neuroscience neuroimaging EEG LSL real-time brain',
    project_urls={
        'Documentation': 'https://bsl-tools.github.io/',
        'Source': 'https://github.com/bsl-tools/bsl',
        'Tracker': 'https://github.com/bsl-tools/bsl/issues'
        },
    platforms='any',
    python_requires='>=3.6',
    install_requires=install_requires,
    extras_require=extras_require,
    packages=find_packages(),
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'bsl = bsl.commands.main:run',
            'bsl_stream_player = bsl.commands.bsl_stream_player:run',
            'bsl_stream_recorder = bsl.commands.bsl_stream_recorder:run',
            'bsl_stream_viewer = bsl.commands.bsl_stream_viewer:run'
          ]
        }
    )



"""
Optional dependency import.
Inspired from pandas: https://pandas.pydata.org/
"""
import importlib


# A mapping from import name to package name (on PyPI) when the package name
# is different.
INSTALL_MAPPING = {
    "serial": "pyserial"
    }


def import_optional_dependency(
        name: str,
        extra: str = "",
        raise_error: bool = True):
    """
    Import an optional dependency.
    By default, if a dependency is missing an ImportError with a nice message
    will be raised.

    Parameters
    ----------
    name : str
        The module name.
    extra : str
        Additional text to include in the ImportError message.
    raise_error : bool
        What to do when a dependency is not found.
        * True : Raise an ImportError.
        * False: If the module is not installed, return None, otherwise, return
        the module.

    Returns
    -------
    maybe_module : Optional[ModuleType]
        The imported module when found.
        None is returned when the package is not found and raise_error is
        False.
    """
    package_name = INSTALL_MAPPING.get(name)
    install_name = package_name if package_name is not None else name

    msg = (
        f"Missing optional dependency '{install_name}'. {extra} "
        f"Use pip or conda to install {install_name}.")

    try:
        module = importlib.import_module(name)
    except ImportError:
        if raise_error:
            raise ImportError(msg) from None
        else:
            return None

    return module
